from . import SendSlackMessage


class SendSlackMessageSpiderStarted(SendSlackMessage):
    message_template = "slack/spider/notifier/start/message.jinja"
    include_attachments = False


class SendSlackMessageSpiderFinished(SendSlackMessage):
    message_template = "slack/spider/notifier/finish/message.jinja"
    attachments_template = "slack/spider/notifier/finish/attachments.jinja"
    include_ok_attachments = False
    include_error_attachments = True
    include_report_link = True
    report_index = 0

    def __init__(
        self,
        include_ok_attachments=None,
        include_error_attachments=None,
        include_report_link=None,
        report_index=None,
        *args,
        **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.include_ok_attachments = (
            include_ok_attachments or self.include_ok_attachments
        )
        self.include_error_attachments = (
            include_error_attachments or self.include_error_attachments
        )
        self.include_report_link = include_report_link or self.include_report_link
        self.report_index = report_index or self.report_index

    @classmethod
    def from_crawler_kwargs(cls, crawler):
        kwargs = super().from_crawler_kwargs(crawler)
        kwargs.update(
            {
                "include_ok_attachments": crawler.settings.get(
                    "SPIDERMON_SLACK_NOTIFIER_INCLUDE_OK_ATTACHMENTS"
                ),
                "include_error_attachments": crawler.settings.get(
                    "SPIDERMON_SLACK_NOTIFIER_INCLUDE_ERROR_ATTACHMENTS"
                ),
                "include_report_link": crawler.settings.get(
                    "SPIDERMON_SLACK_NOTIFIER_INCLUDE_REPORT_LINK"
                ),
                "report_index": crawler.settings.get(
                    "SPIDERMON_SLACK_NOTIFIER_REPORT_INDEX"
                ),
            }
        )
        return kwargs

    def get_attachments(self):
        if (self.monitors_failed and self.include_error_attachments) or (
            self.monitors_passed and self.include_ok_attachments
        ):
            return super().get_attachments()
        else:
            return None

    def get_template_context(self):
        context = super().get_template_context()
        context.update(
            {
                "include_ok_attachments": self.include_ok_attachments,
                "include_error_attachments": self.include_error_attachments,
                "include_report_link": self.include_report_link,
                "report_index": self.report_index,
            }
        )
        return context


class SendSlackMessageSpiderRunning(SendSlackMessageSpiderFinished):
    message_template = "slack/spider/notifier/periodic/message.jinja"
    attachments_template = "slack/spider/notifier/periodic/attachments.jinja"



import logging

from slugify import slugify

logger = logging.getLogger(__name__)

from sentry_sdk import configure_scope
from sentry_sdk.client import Client

from spidermon import Action
from spidermon.exceptions import NotConfigured


class SendSentryMessage(Action):
    sentry_dsn = None
    fake = False
    sentry_log_level = "error"
    project_name = ""
    environment = "Development"

    def __init__(
        self,
        sentry_dsn=None,
        fake=None,
        sentry_log_level=None,
        project_name="",
        environment="",
    ):
        super().__init__()
        self.fake = fake or self.fake
        self.sentry_log_level = sentry_log_level or self.sentry_log_level
        self.sentry_dsn = sentry_dsn or self.sentry_dsn

        self.project_name = project_name or self.project_name
        self.environment = environment or self.environment

        if not self.fake and not self.sentry_dsn:
            raise NotConfigured("Missing SPIDERMON_SENTRY_DSN setting")

        if not self.project_name:
            raise NotConfigured("Missing SPIDERMON_SENTRY_PROJECT_NAME setting")

    @classmethod
    def from_crawler_kwargs(cls, crawler):
        return {
            "fake": crawler.settings.getbool("SPIDERMON_SENTRY_FAKE"),
            "sentry_dsn": crawler.settings.get("SPIDERMON_SENTRY_DSN"),
            "sentry_log_level": crawler.settings.get("SPIDERMON_SENTRY_LOG_LEVEL"),
            "project_name": crawler.settings.get("SPIDERMON_SENTRY_PROJECT_NAME"),
            "environment": crawler.settings.get("SPIDERMON_SENTRY_ENVIRONMENT_TYPE"),
        }

    def run_action(self):
        message = self.get_message()
        if self.fake:
            logger.info(message)
        else:
            self.send_message(message)

    def get_title(self):
        return (
            "{project_name} | {environment} | Spider {spider_name} notification".format(
                project_name=self.project_name,
                environment=self.environment,
                spider_name=self.data.spider.name,
            )
        )

    def get_message(self):
        """
        Returns message dictionary
        """
        message = dict()

        message["title"] = self.get_title()
        if self.data.job:
            message["job_link"] = "https://app.zyte.com/p/{job_id}".format(
                job_id=self.data.job.key
            )

        if self.data.spider:
            message["spider_name"] = self.data.spider.name

        if self.data.stats:
            message["items_count"] = self.data.stats.get("item_scraped_count", 0)

        if self.result:
            message["passed_monitors_count"] = len(self.result.monitors_passed_results)
            message["failed_monitors_count"] = len(self.result.monitors_failed_results)

            failed_monitors = []
            failure_reasons = []

            for result in self.result.monitors_failed_results:
                failed_monitors.append(result.monitor.name)
                failure_reasons.append(result.error)

            message["failure_reasons"] = "\n".join(failure_reasons)
            message["failed_monitors"] = failed_monitors

        return message

    def get_tags(self, message):
        tags = {
            "spider_name": message.get("spider_name", ""),
            "project_name": self.project_name,
        }
        for failed_monitor in message.get("failed_monitors", []):
            key = slugify(failed_monitor.split("/")[-1], max_length=32, separator="_")
            tags[key] = 1
        return tags

    def send_message(self, message):

        sentry_client = Client(dsn=self.sentry_dsn, environment=self.environment)

        with configure_scope() as scope:
            tags = self.get_tags(message)
            for key, val in tags.items():
                scope.set_tag(key, val)

            scope.set_extra("job_link", message.get("job_link", ""))
            scope.set_extra("spider_name", message.get("spider_name", ""))
            scope.set_extra("items_count", message.get("items_count", 0))

            scope.set_extra(
                "passed_monitors_count", message.get("passed_monitors_count", 0)
            )
            scope.set_extra(
                "failed_monitors_count", message.get("failed_monitors_count", 0)
            )

            scope.set_extra("failed_monitors", message.get("failed_monitors", []))

            sentry_client.capture_event(
                {
                    "message": "{title} \n {description}".format(
                        title=message.get("title"),
                        description=message.get("failure_reasons", ""),
                    ),
                    "level": self.sentry_log_level,
                },
                scope=scope,
            )
        logger.info("Notification sent to the sentry dashboard!!")

        sentry_client.close()



import pytest
from scrapy import Item, signals
from scrapy.spiders import Spider
from scrapy.utils.test import get_crawler


class TestItem(Item):

    __test__ = False


@pytest.fixture
def spider():
    settings = {
        "SPIDERMON_ENABLED": True,
        "EXTENSIONS": {"spidermon.contrib.scrapy.extensions.Spidermon": 100},
        "SPIDERMON_ADD_FIELD_COVERAGE": True,
    }
    crawler = get_crawler(settings_dict=settings)

    spider = Spider.from_crawler(crawler, "example.com")

    return spider


def test_add_stats_item_scraped_count_by_item_type(spider):
    for _ in range(15):
        spider.crawler.signals.send_catch_log_deferred(
            signal=signals.item_scraped,
            item={"_type": "regular_dict"},
            response="",
            spider=spider,
        )

    for _ in range(20):
        spider.crawler.signals.send_catch_log_deferred(
            signal=signals.item_scraped,
            item=Item(),
            response="",
            spider=spider,
        )

    for _ in range(25):
        spider.crawler.signals.send_catch_log_deferred(
            signal=signals.item_scraped,
            item=TestItem(),
            response="",
            spider=spider,
        )

    stats = spider.crawler.stats.get_stats()

    assert stats.get("spidermon_item_scraped_count") == 60
    assert stats.get("spidermon_item_scraped_count/dict") == 15
    assert stats.get("spidermon_item_scraped_count/Item") == 20
    assert stats.get("spidermon_item_scraped_count/TestItem") == 25


def test_item_scraped_count_single_field(spider):
    returned_items = [{"field1": "value1"}]

    for item in returned_items:
        spider.crawler.signals.send_catch_log_deferred(
            signal=signals.item_scraped,
            item=item,
            response="",
            spider=spider,
        )

    stats = spider.crawler.stats.get_stats()
    assert stats.get("spidermon_item_scraped_count/dict/field1") == 1


def test_item_scraped_count_multiple_field(spider):
    returned_items = [{"field1": "value1", "field2": "value2"}]

    for item in returned_items:
        spider.crawler.signals.send_catch_log_deferred(
            signal=signals.item_scraped,
            item=item,
            response="",
            spider=spider,
        )

    stats = spider.crawler.stats.get_stats()
    assert stats.get("spidermon_item_scraped_count/dict/field1") == 1
    assert stats.get("spidermon_item_scraped_count/dict/field2") == 1


def test_item_scraped_count_multiple_items(spider):
    returned_items = [
        {"field1": "value1", "field2": "value2"},
        {"field1": "value1", "field2": "value2"},
    ]

    for item in returned_items:
        spider.crawler.signals.send_catch_log_deferred(
            signal=signals.item_scraped,
            item=item,
            response="",
            spider=spider,
        )

    stats = spider.crawler.stats.get_stats()
    assert stats.get("spidermon_item_scraped_count/dict/field1") == 2
    assert stats.get("spidermon_item_scraped_count/dict/field2") == 2


def test_item_scraped_count_multiple_items_field_missing(spider):
    returned_items = [
        {"field1": "value1", "field2": "value2"},
        {
            "field1": "value1",
        },
    ]

    for item in returned_items:
        spider.crawler.signals.send_catch_log_deferred(
            signal=signals.item_scraped,
            item=item,
            response="",
            spider=spider,
        )

    stats = spider.crawler.stats.get_stats()
    assert stats.get("spidermon_item_scraped_count/dict/field1") == 2
    assert stats.get("spidermon_item_scraped_count/dict/field2") == 1


def test_item_scraped_count_single_nested_field(spider):
    returned_items = [{"field1": {"field1.1": "value1.1"}}]

    for item in returned_items:
        spider.crawler.signals.send_catch_log_deferred(
            signal=signals.item_scraped,
            item=item,
            response="",
            spider=spider,
        )

    stats = spider.crawler.stats.get_stats()
    assert stats.get("spidermon_item_scraped_count/dict") == 1
    assert stats.get("spidermon_item_scraped_count/dict/field1") == 1
    assert stats.get("spidermon_item_scraped_count/dict/field1/field1.1") == 1


def test_item_scraped_count_multiple_nested_field(spider):
    returned_items = [
        {
            "field1": {"field1.1": "value1.1"},
            "field2": "value2",
            "field3": {"field3.1": "value3.1"},
        },
        {
            "field1": {
                "field1.1": "value1.1",
                "field1.2": "value1.2",
            },
            "field2": "value2",
        },
    ]

    for item in returned_items:
        spider.crawler.signals.send_catch_log_deferred(
            signal=signals.item_scraped,
            item=item,
            response="",
            spider=spider,
        )

    stats = spider.crawler.stats.get_stats()

    assert stats.get("spidermon_item_scraped_count/dict") == 2
    assert stats.get("spidermon_item_scraped_count/dict/field1") == 2
    assert stats.get("spidermon_item_scraped_count/dict/field1/field1.1") == 2
    assert stats.get("spidermon_item_scraped_count/dict/field1/field1.2") == 1
    assert stats.get("spidermon_item_scraped_count/dict/field2") == 2
    assert stats.get("spidermon_item_scraped_count/dict/field3") == 1


def test_item_scraped_count_with_slash_on_field_name(spider):
    returned_items = [{"field1/with/slash": "value1", "field2": "value2"}]

    for item in returned_items:
        spider.crawler.signals.send_catch_log_deferred(
            signal=signals.item_scraped,
            item=item,
            response="",
            spider=spider,
        )

    stats = spider.crawler.stats.get_stats()
    assert stats.get("spidermon_item_scraped_count/dict/field1/with/slash") == 1
    assert stats.get("spidermon_item_scraped_count/dict/field2") == 1


def test_do_not_add_field_coverage_when_spider_closes_if_do_not_have_field_coverage_settings():
    settings = {
        "SPIDERMON_ENABLED": True,
        "EXTENSIONS": {"spidermon.contrib.scrapy.extensions.Spidermon": 100},
        "SPIDERMON_ADD_FIELD_COVERAGE": False,
    }
    crawler = get_crawler(settings_dict=settings)
    spider = Spider.from_crawler(crawler, "example.com")

    item = {"field1": "value1"}
    spider.crawler.signals.send_catch_log_deferred(
        signal=signals.item_scraped,
        item=item,
        response="",
        spider=spider,
    )  # Return item to have some stats to calculate coverage

    crawler.signals.send_catch_log(
        signal=signals.spider_closed, spider=spider, reason=None
    )

    stats = spider.crawler.stats.get_stats()

    assert stats.get("spidermon_field_coverage/dict/field1") is None


def test_add_field_coverage_when_spider_closes_if_have_field_coverage_settings():
    settings = {
        "SPIDERMON_ENABLED": True,
        "EXTENSIONS": {"spidermon.contrib.scrapy.extensions.Spidermon": 100},
        "SPIDERMON_ADD_FIELD_COVERAGE": True,
    }
    crawler = get_crawler(settings_dict=settings)
    spider = Spider.from_crawler(crawler, "example.com")

    item = {"field1": "value1"}
    spider.crawler.signals.send_catch_log_deferred(
        signal=signals.item_scraped,
        item=item,
        response="",
        spider=spider,
    )  # Return item to have some stats to calculate coverage

    crawler.signals.send_catch_log(
        signal=signals.spider_closed, spider=spider, reason=None
    )

    stats = spider.crawler.stats.get_stats()

    assert stats.get("spidermon_field_coverage/dict/field1") == 1.0



from spidermon import monitors, Monitor, MonitorSuite


# ----------------------------------
# Monitors
# ----------------------------------
class UnnamedMonitor(Monitor):
    def test_without_name(self):
        pass

    @monitors.name("A Test")
    def test_with_name(self):
        pass


@monitors.name("Class Monitor")
class NamedMonitor(Monitor):
    def test_without_name(self):
        pass

    @monitors.name("A Test")
    def test_with_name(self):
        pass


# ----------------------------------
# Child Suites
# ----------------------------------
class BaseChildSuite(MonitorSuite):
    monitors = [
        UnnamedMonitor,
        NamedMonitor,
        ("Instance Monitor", UnnamedMonitor),
        ("Instance Monitor", NamedMonitor),
    ]


class ChildUnnamedSuite(BaseChildSuite):
    pass


@monitors.name("The Child Suite")
class ChildNamedSuite(BaseChildSuite):
    pass


# ----------------------------------
# Top Suites
# ----------------------------------
class BaseTopSuite(MonitorSuite):
    monitors = [
        ChildUnnamedSuite,
        ChildNamedSuite,
        ("Instance Suite Name", ChildUnnamedSuite),
        ("Instance Suite Name", ChildNamedSuite),
    ]


class UnnamedTopSuite(BaseTopSuite):
    pass


@monitors.name("The Top Suite")
class NamedTopSuite(BaseTopSuite):
    pass



from spidermon.contrib.validation.translator import MessageTranslator
from spidermon.contrib.validation import messages


class SchematicsMessageTranslator(MessageTranslator):
    messages = {
        r"^Rogue field$": messages.UNEXPECTED_FIELD,
        # BaseType
        r"^This field is required.$": messages.MISSING_REQUIRED_FIELD,
        r"^Value \(.*?\) must be one of \[.*?\]\.$": messages.VALUE_NOT_IN_CHOICES,
        # StringType
        r"^Couldn't interpret '.*' as string\.$": messages.INVALID_STRING,
        r"^String value is too long\.$": messages.FIELD_TOO_LONG,
        r"^String value is too short\.$": messages.FIELD_TOO_SHORT,
        r"^String value did not match validation regex\.$": messages.REGEX_NOT_MATCHED,
        # DateTimeType
        r"^Could not parse .+\. Should be ISO ?8601(?: or timestamp)?\.$": messages.INVALID_DATETIME,
        r"^Could not parse .+\. Valid formats: .+$": messages.INVALID_DATETIME,
        # DateType
        r"^Could not parse .+\. Should be ISO ?8601 \(YYYY-MM-DD\)\.$": messages.INVALID_DATE,
        # NumberType
        r"^.+ value should be greater than .+$": messages.NUMBER_TOO_LOW,
        r"^.+ value should be less than .+$": messages.NUMBER_TOO_HIGH,
        # IntType
        r"^Value '.*' is not int\.?$": messages.INVALID_INT,
        # FloatType
        r"^Value '.*' is not float\.?$": messages.INVALID_FLOAT,
        # LongType
        r"^Value '.*' is not long\.?$": messages.INVALID_LONG,
        # Decimalype
        r"^Number '.*' failed to convert to a decimal\.?$": messages.INVALID_DECIMAL,
        r"^Value '.*' is not decimal\.?$": messages.INVALID_DECIMAL,
        r"^Value should be greater than .+$": messages.NUMBER_TOO_LOW,
        r"^Value should be less than .+$": messages.NUMBER_TOO_HIGH,
        # BooleanType
        r"^Must be either true or false\.$": messages.INVALID_BOOLEAN,
        # EmailType
        r"^Not a well[ -]formed email address\.$": messages.INVALID_EMAIL,
        # URLType
        r"^Not a well[ -]formed URL\.$": messages.INVALID_URL,
        # UUIDType
        r"^Couldn't interpret '.*' value as UUID\.$": messages.INVALID_UUID,
        # IPv4Type
        r"^Invalid IPv4 address$": messages.INVALID_IPV4,
        # HashType
        r"^Hash value is wrong length\.$": messages.INVALID_HASH_LENGTH,
        r"^Hash value is not hexadecimal\.$": messages.INVALID_HASH,
        # ListType
        r"^Invalid list$": messages.INVALID_LIST,
        r"^Could not interpret the value as a list$": messages.INVALID_LIST,
        r"^Please provide at least \d+ items?\.$": messages.LIST_TOO_SHORT,
        r"^Please provide no more than \d+ items?\.$": messages.LIST_TOO_LONG,
        # DictType
        r"^Only (?:dictionaries|mappings) may be used in a DictType$": messages.INVALID_DICT,
        # DictType
        r"^Please use a mapping for this field or .+ "
        r"instance instead of .*\.$": messages.INVALID_CHILD_CONTENT,
    }



from unittest import TestCase

from spidermon import settings
from .options import MonitorOptions, MonitorOptionsMetaclass


class Monitor(TestCase, metaclass=MonitorOptionsMetaclass):
    def __init__(self, methodName="runTest", name=None):
        super().__init__(methodName)
        self._name = name
        self._data = None
        self._parent = None
        self._init_method()

    @property
    def name(self):
        return "/".join([self.monitor_name, self.method_name])

    @property
    def full_name(self):
        parts = []
        if self.parent and self.parent.full_name:
            parts.append(self.parent.full_name)
        parts.append(self.name)
        return "/".join(parts)

    @property
    def level(self):
        return (
            self.method_level
            or self.monitor_level
            or self.parent_level
            or settings.MONITOR.LEVELS.DEFAULT
        )

    @property
    def order(self):
        return self.method.options.order

    @property
    def monitor_name(self):
        return self._name or self.options.name or self.__class__.__name__

    @property
    def monitor_full_name(self):
        parts = []
        if self.parent and self.parent.full_name:
            parts.append(self.parent.full_name)
        parts.append(self.monitor_name)
        return "/".join(parts)

    @property
    def monitor_description(self):
        return (
            self.options.description
            or self.__class__.__doc__
            or settings.MONITOR.DEFAULT_DESCRIPTION
        )

    @property
    def monitor_level(self):
        return self.options.level

    @property
    def method(self):
        return getattr(self, self._testMethodName)

    @property
    def method_name(self):
        return self.method.options.name or self._testMethodName

    @property
    def method_description(self):
        return (
            self.method.options.description
            or self.method.__func__.__doc__
            or settings.MONITOR.DEFAULT_DESCRIPTION
        )

    @property
    def method_level(self):
        return self.method.options.level

    @property
    def parent(self):
        return self._parent

    @property
    def parent_level(self):
        return self.parent.level

    def set_parent(self, parent):
        self._parent = parent

    def init_data(self, data):
        self.data = data

    def debug_tree(self, level=0):
        return level * "\t" + repr(self) + "\n"

    def _init_method(self):
        MonitorOptions.add_or_create(self.method.__func__)

    def __repr__(self):
        return "<MONITOR:({}) at {}>".format(self.name, hex(id(self)))

    def __str__(self):
        return self.name

    def __hash__(self):
        return hash(id(self))



MONITOR_SCHEMA = {
    "type": "object",
    "properties": {
        "name": {"type": "string", "minLength": 1},
        "description": {"type": "string", "minLength": 1},
        "tests": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "name": {"type": "string", "minLength": 1},
                    "description": {"type": "string", "minLength": 1},
                    "expression": {"type": "string", "minLength": 1},
                    "fail_reason": {"type": "string", "minLength": 1},
                },
                "required": ["name", "expression"],
            },
        },
    },
    "required": ["name", "tests"],
}



from .exceptions import InvalidDataOperation


class Data(dict):
    """
    Immutable dict class with attribute access.

    example:
    >> s = Data({'scraped_items': 100})
    >> s['scraped_items']
    100
    >> s.scraped_items
    100
    """

    def __getattr__(self, name):
        if name in self:
            return self[name]
        else:
            raise AttributeError("Key '%s' not found." % name)

    def _immutable(self, *args, **kws):
        raise InvalidDataOperation(
            "Immutable Data! You cannot add or modify read-only data."
        )

    update = _immutable
    setdefault = _immutable
    clear = _immutable
    pop = _immutable
    popitem = _immutable
    __setitem__ = _immutable
    __delitem__ = _immutable
    __setattr__ = _immutable



import json
from jsonschema import validate

from spidermon import Monitor
from spidermon.core.options import MonitorOptions
from spidermon.python import Interpreter, schemas
from spidermon.exceptions import InvalidMonitor, NotConfigured
from spidermon import settings


class PythonExpressionsMonitor(Monitor):
    _classes_counter = 0
    _test_methods_counter = 0
    _test_methods_prefix = "test"

    @classmethod
    def generate_class_name(cls):
        cls._classes_counter += 1
        return "%s%d" % (cls.__name__, cls._classes_counter)

    @classmethod
    def generate_method_name(cls):
        cls._test_methods_counter += 1
        return "%s_python_expression_%d" % (
            cls._test_methods_prefix,
            cls._test_methods_counter,
        )

    def get_context_data(self):
        raise NotConfigured("Context data needs to be set up")


def create_monitor_class_from_json(monitor_json, monitor_class=None):
    monitor_dict = json.loads(monitor_json)
    validate(monitor_dict, schemas.MONITOR_SCHEMA)
    return create_monitor_class_from_dict(monitor_dict, monitor_class)


def create_monitor_class_from_dict(monitor_dict, monitor_class=None):
    tests = []
    for test in monitor_dict.get("tests", []):
        tests.append(
            (
                test["expression"],
                test.get("name", None),
                test.get("description", None),
                test.get("fail_message", None),
            )
        )
    klass = _create_monitor_class(tests, monitor_class)
    klass.options.name = monitor_dict.get("name", settings.MONITOR.DEFAULT_NAME)
    klass.options.description = monitor_dict.get(
        "description", settings.MONITOR.DEFAULT_DESCRIPTION
    )
    return klass


def _create_monitor_class(expressions, monitor_class=None):
    monitor_class = monitor_class or PythonExpressionsMonitor
    if not issubclass(monitor_class, PythonExpressionsMonitor):
        msg = "Python expressions monitors must subclass PythonExpressionsMonitor"
        raise InvalidMonitor(msg)
    klass = type(monitor_class.generate_class_name(), (monitor_class,), {})
    for e in expressions:
        if isinstance(e, tuple):
            method = _create_test_method(*e)
        else:
            method = _create_test_method(e)
        setattr(klass, monitor_class.generate_method_name(), method)
    return klass


def _create_test_method(expression, name=None, description=None, fail_reason=None):
    def _test_method(self):
        interpreter = Interpreter()
        context = self.get_context_data()
        result = interpreter.eval(expression, context=context)
        if result is not None:
            self.assertTrue(
                bool(result),
                msg=('Expression not safisfied: "%s"' % expression)
                if not fail_reason
                else interpreter.eval(fail_reason, context=context),
            )

    test_method = _test_method
    MonitorOptions.add_or_create(test_method)
    test_method.options.name = name or settings.MONITOR.DEFAULT_NAME
    test_method.options.description = (
        description or settings.MONITOR.DEFAULT_DESCRIPTION
    )
    return test_method



import re

from jsonschema.validators import validator_for
from spidermon.contrib.validation.validator import Validator

from .translator import JSONSchemaMessageTranslator
from .formats import format_checker


REQUIRED_RE = re.compile("'(.+)' is a required property")


class JSONSchemaValidator(Validator):
    default_translator = JSONSchemaMessageTranslator()
    name = "JSONSchema"

    def __init__(self, schema, translator=None, use_default_translator=True):
        super().__init__(
            translator=translator, use_default_translator=use_default_translator
        )
        self._schema = schema

    def _validate(self, data, strict=False):
        validator_cls = validator_for(self._schema)
        validator = validator_cls(schema=self._schema, format_checker=format_checker)
        errors = validator.iter_errors(data)

        for error in errors:
            absolute_path = list(error.absolute_path)
            required_match = REQUIRED_RE.search(error.message)
            if required_match:
                absolute_path.append(required_match.group(1))
            field_name = ".".join([str(p) for p in absolute_path])
            self._add_errors({field_name: [error.message]})



import pytest

from spidermon import MonitorLoader
from spidermon.exceptions import InvalidMonitor

from .fixtures.cases import *
from .test_suites import check_suite


@pytest.fixture
def loader():
    return MonitorLoader()


def test_loading(loader):
    check_suite(
        suite=loader.load_suite_from_monitor(EmptyMonitor),
        expected_number_of_monitors=0,
    )
    check_suite(
        suite=loader.load_suite_from_monitor(Monitor01), expected_number_of_monitors=3
    )
    check_suite(
        suite=loader.load_suite_from_monitor(Monitor02), expected_number_of_monitors=2
    )


def test_loading_errors(loader):
    with pytest.raises(InvalidMonitor):
        loader.load_suite_from_monitor(None)
    with pytest.raises(InvalidMonitor):
        loader.load_suite_from_monitor(10)
    with pytest.raises(InvalidMonitor):
        loader.load_suite_from_monitor(object)


def test_testcase_names(loader):
    _check_testcase_names(loader=loader, monitor_class=EmptyMonitor, expected_names=[])
    _check_testcase_names(
        loader=loader,
        monitor_class=Monitor01,
        expected_names=["test_a", "test_b", "test_c"],
    )
    _check_testcase_names(
        loader=loader, monitor_class=Monitor02, expected_names=["test_d", "test_e"]
    )


def _check_testcase_names(loader, monitor_class, expected_names):
    names = loader.get_testcase_names(monitor_class)
    assert names == expected_names



import re
from collections import defaultdict


RE_PATTERN_INSTANCE = type(re.compile(""))


class Validator:
    default_translator = None
    name = "validator"

    def __init__(self, translator=None, use_default_translator=True):
        self._errors = defaultdict(list)
        if not translator and use_default_translator and self.default_translator:
            translator = self.default_translator
        self._translator = translator

    def validate(self, data, strict=True):
        self._reset()
        self._validate(data, strict=strict)
        return not self.has_errors, self.errors

    def _reset(self):
        self._errors = defaultdict(list)

    def _validate(self, data, strict=True):
        raise NotImplementedError

    @property
    def has_errors(self):
        return len(self._errors) > 0

    @property
    def errors(self):
        if not self._translator:
            return self._errors
        else:
            return {
                field_name: self._translator.translate_messages(messages)
                for field_name, messages in self._errors.items()
            }

    def _add_errors(self, errors):
        for field_name, messages in errors.items():
            self._errors[field_name] += (
                messages if isinstance(messages, list) else [messages]
            )



import hashlib

from boto.s3.connection import S3Connection
from boto.s3.key import Key

from spidermon.exceptions import NotConfigured

from . import CreateReport


DEFAULT_S3_REGION_ENDPOINT = "s3.amazonaws.com"
DEFAULT_S3_CONTENT_TYPE = "text/html"
URL_SECRET_KEY = "The secret to life the universe and everything"


class S3Uploader:
    def __init__(self, aws_key, aws_secret):
        self.connection = S3Connection(
            aws_access_key_id=aws_key, aws_secret_access_key=aws_secret
        )

    def upload_from_file(
        self, source_filename, s3_bucket, s3_filename, headers=None, make_public=False
    ):
        with open(source_filename) as f:
            self._upload_with_method(
                bucket=s3_bucket,
                method_name="set_contents_from_file",
                filename=s3_filename,
                content=f,
                headers=headers,
                make_public=make_public,
            )

    def upload_from_content(
        self, content, s3_bucket, s3_filename, headers=None, make_public=False
    ):
        self._upload_with_method(
            bucket=s3_bucket,
            method_name="set_contents_from_string",
            filename=s3_filename,
            content=content,
            headers=headers,
            make_public=make_public,
        )

    def _upload_with_method(
        self, bucket, method_name, filename, content, headers=None, make_public=False
    ):
        # Get bucket without validation (Needed to be used with credentials w/o listing perms)
        bucket = self.connection.get_bucket(bucket, validate=False)
        f = Key(bucket)
        f.key = filename
        getattr(f, method_name)(content, headers=headers)
        if make_public:
            f.make_public()


class CreateS3Report(CreateReport):
    aws_access_key = None
    aws_secret_key = None
    s3_bucket = None
    s3_filename = None
    s3_region_endpoint = DEFAULT_S3_REGION_ENDPOINT
    make_public = True
    content_type = DEFAULT_S3_CONTENT_TYPE

    def __init__(
        self,
        aws_access_key=None,
        aws_secret_key=None,
        s3_bucket=None,
        s3_filename=None,
        s3_region_endpoint=None,
        make_public=False,
        content_type=None,
        *args,
        **kwargs
    ):
        super().__init__(*args, **kwargs)

        self.aws_access_key = aws_access_key or self.aws_access_key
        self.aws_secret_key = aws_secret_key or self.aws_secret_key
        self.s3_bucket = s3_bucket or self.s3_bucket
        self.s3_region_endpoint = s3_region_endpoint or self.s3_region_endpoint
        self.s3_filename = s3_filename or self.s3_filename
        self.make_public = make_public or self.make_public
        self.content_type = content_type or self.content_type
        if not self.aws_access_key:
            raise NotConfigured("You must provide the AWS Access Key.")
        if not self.aws_secret_key:
            raise NotConfigured("You must provide the AWS Secret Key.")
        if not self.s3_bucket:
            raise NotConfigured("You must define the s3 bucket.")
        if not self.s3_filename:
            raise NotConfigured("You must define the s3 filename.")

    @classmethod
    def from_crawler_kwargs(cls, crawler):
        kwargs = super().from_crawler_kwargs(crawler)
        (aws_access_key_id, aws_secret_access_key) = get_aws_credentials(
            crawler.settings
        )
        kwargs.update(
            {
                "aws_access_key": aws_access_key_id,
                "aws_secret_key": aws_secret_access_key,
                "s3_bucket": crawler.settings.get("SPIDERMON_REPORT_S3_BUCKET"),
                "s3_filename": crawler.settings.get("SPIDERMON_REPORT_S3_FILENAME"),
                "s3_region_endpoint": crawler.settings.get(
                    "SPIDERMON_REPORT_S3_REGION_ENDPOINT"
                ),
                "make_public": crawler.settings.get("SPIDERMON_REPORT_S3_MAKE_PUBLIC"),
                "content_type": crawler.settings.get(
                    "SPIDERMON_REPORT_S3_CONTENT_TYPE"
                ),
            }
        )
        return kwargs

    def after_render_report(self):
        s3 = S3Uploader(self.aws_access_key, self.aws_secret_key)
        s3.upload_from_content(
            content=self.report,
            s3_bucket=self.s3_bucket,
            s3_filename=self.get_s3_filename(),
            headers={"Content-Type": self.content_type},
            make_public=self.make_public,
        )

    def get_s3_filename(self):
        return "reports/{secret}/{filename}".format(
            secret=self.get_url_secret(),
            filename=self.render_text_template(self.s3_filename),
        )

    def get_s3_report_url(self):
        return "https://{region}/{bucket}/{filename}".format(
            region=self.s3_region_endpoint,
            bucket=self.s3_bucket,
            filename=self.get_s3_filename(),
        )

    def get_url_secret(self):
        secret = URL_SECRET_KEY
        if self.data.job:
            secret += str(self.data.job.key.split("/")[0])
        return hashlib.md5(secret.encode()).hexdigest()

    def get_meta(self):
        report_url = self.get_s3_report_url()
        return {"reports_links": self.data.meta.get("reports", []) + [report_url]}



import ast

from spidermon.exceptions import InvalidExpression


class Interpreter:

    ast_allowed_nodes = (
        "expr",
        "name",
        "load",
        "call",
        "store",
        "str",
        "unicode",
        "num",
        "list",
        "dict",
        "set",
        "tuple",  # Data types
        "constant",
        "nameconstant",  # language constant (None)
        "unaryop",
        "usub",  # Unary arithmetic operators
        "binop",
        "add",
        "sub",
        "div",
        "mult",
        "mod",
        "pow",
        "floordiv",  # Binary arithmetic operators
        "compare",
        "eq",
        "noteq",
        "gt",
        "lt",
        "gte",
        "lte",  # Comparison operators
        "bitand",
        "bitor",
        "bitxor",
        "invert",
        "lshift",
        "rshift",  # Bitwise operators
        "boolop",
        "and",
        "or",
        "not",  # Logical operators
        "in",
        "notin",  # Membership operators
        "is",
        "isnot",  # Identity operators
        "ifexp",  # Inline if statement
        "subscript",
        "index",
        "slice",
        "extslice",  # Subscripting
        "listcomp",
        "setcomp",
        "dictcomp",
        "generatorexp",
        "comprehension",  # Comprehensions
        "attribute",  # Attribute access
    )

    allowed_objects = (
        str,
        str,  # strings
        int,
        float,
        complex,  # numbers
        list,
        dict,
        set,
        tuple,  # sequences
        type(None),
        bool,  # others
    )

    def check(self, expression):
        if not isinstance(expression, str):
            raise InvalidExpression("Python expressions must be defined as strings")
        if not expression:
            raise InvalidExpression("Empty python expression")

        try:
            tree = ast.parse(expression)
        except SyntaxError as e:
            raise e

        if not tree.body:
            raise InvalidExpression("Empty python expression")
        elif len(tree.body) > 1:
            raise InvalidExpression(
                "Python expressions must be a single line expression"
            )

        start_node = tree.body[0]
        if not isinstance(start_node, ast.Expr):
            raise InvalidExpression(
                "Python string must be an expression: '%s' found"
                % start_node.__class__.__name__
            )

        self._check_node(start_node)

    def eval(self, expression, context=None, check=True):
        if check:
            self.check(expression)
        return eval(expression, context)

    def _check_node(self, node):
        if isinstance(node, list):
            self._check_node_list(node)
        elif isinstance(node, ast.AST):
            if not self._is_allowed_ast_node(node):
                self._raise_not_allowed_node(node)
            self._check_node_fields(node)
        elif not isinstance(node, self.allowed_objects):
            self._raise_not_allowed_node(node)

    def _check_node_list(self, node_list):
        for node in node_list:
            self._check_node(node)

    def _check_node_fields(self, node):
        for field in [f for _, f in ast.iter_fields(node)]:
            self._check_node(field)

    def _is_allowed_ast_node(self, node):
        return node.__class__.__name__.lower() in self.ast_allowed_nodes

    def _raise_not_allowed_node(self, node):
        raise InvalidExpression(
            "'%s' definition not allowed in python expressions"
            % node.__class__.__name__
        )



from slugify import slugify


STATS_DEFAULT_VALIDATION_PREFIX = "spidermon/validation"


class NAMES:
    ITEMS = "items"
    DROPPED = "dropped"
    FIELDS = "fields"
    ERRORS = "errors"
    VALIDATORS = "validators"


class ValidationStatsManager:
    def __init__(self, stats, prefix=None, slugify=True):
        self.stats = stats
        self.prefix = prefix or STATS_DEFAULT_VALIDATION_PREFIX
        self.slugify = slugify

    def add_validator(self, type, class_name):
        self.stats.inc_value(self._get_stats_name(NAMES.VALIDATORS))
        self.stats.set_value(
            self._get_stats_name(NAMES.VALIDATORS, type, class_name), True
        )

    def add_field_error(self, field, error):
        self.stats.inc_value(self._get_stats_name(NAMES.FIELDS, NAMES.ERRORS))
        self.stats.inc_value(self._get_stats_name(NAMES.FIELDS, NAMES.ERRORS, error))
        self.stats.inc_value(
            self._get_stats_name(NAMES.FIELDS, NAMES.ERRORS, error) + "/" + field
        )

    def add_fields(self, count):
        self.stats.inc_value(self._get_stats_name(NAMES.FIELDS), count=count)

    def add_item(self):
        self.stats.inc_value(self._get_stats_name(NAMES.ITEMS))

    def add_dropped_item(self):
        self.stats.inc_value(self._get_stats_name(NAMES.ITEMS, NAMES.DROPPED))

    def add_item_with_errors(self):
        self.stats.inc_value(self._get_stats_name(NAMES.ITEMS, NAMES.ERRORS))

    def _get_stats_name(self, *names):
        return "/".join([self.prefix] + list([self._get_name(n) for n in names]))

    def _get_name(self, name):
        return slugify(text=name, separator="_").lower() if self.slugify else name



from scrapy import signals
from scrapy.exceptions import NotConfigured
from scrapy.utils.misc import load_object
from twisted.internet.task import LoopingCall

from spidermon import MonitorSuite
from spidermon.contrib.scrapy.runners import SpiderMonitorRunner
from spidermon.python import factory
from spidermon.python.monitors import ExpressionsMonitor
from spidermon.utils.field_coverage import calculate_field_coverage
from spidermon.utils.hubstorage import hs


class Spidermon:
    def __init__(
        self,
        crawler,
        spider_opened_suites=None,
        spider_closed_suites=None,
        engine_stopped_suites=None,
        spider_opened_expression_suites=None,
        spider_closed_expression_suites=None,
        engine_stopped_expression_suites=None,
        expressions_monitor_class=None,
        periodic_suites=None,
    ):
        if not crawler.settings.getbool("SPIDERMON_ENABLED"):
            raise NotConfigured
        self.crawler = crawler

        self.spider_opened_suites = [
            self.load_suite(s) for s in spider_opened_suites or []
        ]
        self.spider_opened_suites += [
            self.load_expression_suite(s, expressions_monitor_class)
            for s in spider_opened_expression_suites or []
        ]

        self.spider_closed_suites = [
            self.load_suite(s) for s in spider_closed_suites or []
        ]
        self.spider_closed_suites += [
            self.load_expression_suite(s, expressions_monitor_class)
            for s in spider_closed_expression_suites or []
        ]

        self.engine_stopped_suites = [
            self.load_suite(s) for s in engine_stopped_suites or []
        ]
        self.engine_stopped_suites += [
            self.load_expression_suite(s, expressions_monitor_class)
            for s in engine_stopped_expression_suites or []
        ]

        self.periodic_suites = periodic_suites or {}
        self.periodic_tasks = {}

    def load_suite(self, suite_to_load):
        try:
            suite_class = load_object(suite_to_load)
        except Exception as e:
            raise e  # TO-DO
        if not issubclass(suite_class, MonitorSuite):
            raise Exception  # TO-DO
        return suite_class(crawler=self.crawler)

    def load_expression_suite(self, suite_to_load, monitor_class=None):
        if monitor_class:
            monitor_class = load_object(monitor_class)
        else:
            monitor_class = ExpressionsMonitor
        monitor = factory.create_monitor_class_from_dict(
            monitor_dict=suite_to_load, monitor_class=monitor_class
        )
        suite = MonitorSuite(crawler=self.crawler)
        suite.add_monitor(monitor)
        return suite

    @classmethod
    def from_crawler(cls, crawler):
        ext = cls(
            crawler=crawler,
            spider_opened_suites=crawler.settings.getlist(
                "SPIDERMON_SPIDER_OPEN_MONITORS"
            ),
            spider_closed_suites=crawler.settings.getlist(
                "SPIDERMON_SPIDER_CLOSE_MONITORS"
            ),
            engine_stopped_suites=crawler.settings.getlist(
                "SPIDERMON_ENGINE_STOP_MONITORS"
            ),
            spider_opened_expression_suites=crawler.settings.getlist(
                "SPIDERMON_SPIDER_OPEN_EXPRESSION_MONITORS"
            ),
            spider_closed_expression_suites=crawler.settings.getlist(
                "SPIDERMON_SPIDER_CLOSE_EXPRESSION_MONITORS"
            ),
            engine_stopped_expression_suites=crawler.settings.getlist(
                "SPIDERMON_ENGINE_STOP_EXPRESSION_MONITORS"
            ),
            expressions_monitor_class=crawler.settings.get(
                "SPIDERMON_EXPRESSIONS_MONITOR_CLASS"
            ),
            periodic_suites=crawler.settings.getdict("SPIDERMON_PERIODIC_MONITORS"),
        )
        crawler.signals.connect(ext.spider_opened, signal=signals.spider_opened)
        crawler.signals.connect(ext.spider_closed, signal=signals.spider_closed)
        crawler.signals.connect(ext.engine_stopped, signal=signals.engine_stopped)

        has_field_coverage = crawler.settings.getbool("SPIDERMON_ADD_FIELD_COVERAGE")
        if has_field_coverage:
            crawler.signals.connect(ext.item_scraped, signal=signals.item_scraped)

        return ext

    def spider_opened(self, spider):
        self._run_suites(spider, self.spider_opened_suites)
        self.periodic_tasks[spider] = []
        for suite, time in self.periodic_suites.items():
            task = LoopingCall(self._run_periodic_suites, spider, [suite])
            self.periodic_tasks[spider].append(task)
            task.start(time, now=False)

    def spider_closed(self, spider):
        self._add_field_coverage_to_stats()

        self._run_suites(spider, self.spider_closed_suites)
        for task in self.periodic_tasks[spider]:
            task.stop()

    def engine_stopped(self):
        spider = self.crawler.spider
        self._run_suites(spider, self.engine_stopped_suites)

    def _count_item(self, item, skip_none_values, item_count_stat=None):
        if item_count_stat is None:
            item_type = type(item).__name__
            item_count_stat = f"spidermon_item_scraped_count/{item_type}"
            self.crawler.stats.inc_value(item_count_stat)

        for field_name, value in item.items():
            if skip_none_values and value is None:
                continue

            field_item_count_stat = f"{item_count_stat}/{field_name}"
            self.crawler.stats.inc_value(field_item_count_stat)

            if isinstance(value, dict):
                self._count_item(value, skip_none_values, field_item_count_stat)
                continue

    def _add_field_coverage_to_stats(self):
        stats = self.crawler.stats.get_stats()
        coverage_stats = calculate_field_coverage(stats)
        stats.update(coverage_stats)

    def item_scraped(self, item, response, spider):
        skip_none_values = spider.crawler.settings.getbool(
            "SPIDERMON_FIELD_COVERAGE_SKIP_NONE", False
        )
        self.crawler.stats.inc_value("spidermon_item_scraped_count")
        self._count_item(item, skip_none_values)

    def _run_periodic_suites(self, spider, suites):
        suites = [self.load_suite(s) for s in suites]
        self._run_suites(spider, suites)

    def _run_suites(self, spider, suites):
        data = self._generate_data_for_spider(spider)
        for suite in suites:
            runner = SpiderMonitorRunner(spider=spider)
            runner.run(suite, **data)

    def _generate_data_for_spider(self, spider):
        return {
            "stats": self.crawler.stats.get_stats(spider),
            "stats_history": spider.stats_history
            if hasattr(spider, "stats_history")
            else [],
            "crawler": self.crawler,
            "spider": spider,
            "job": hs.job if hs.available else None,
        }



from .fixtures.ordering import *

SUITE_SEQUENCES = [
    # ------------------------------------------------------------------------------------------------------------
    # monitor sequence                                      expected sequence
    # ------------------------------------------------------------------------------------------------------------
    # Unordered: should follow same insertion order (FIFO)
    ([Unordered.A, Unordered.B], [Unordered.A, Unordered.B]),
    ([Unordered.B, Unordered.A], [Unordered.B, Unordered.A]),
    ([Unordered.C, Unordered.D], [Unordered.C, Unordered.D]),
    ([Unordered.D, Unordered.C], [Unordered.D, Unordered.C]),
    (
        [Unordered.A, Unordered.B, Unordered.C, Unordered.D],
        [Unordered.A, Unordered.B, Unordered.C, Unordered.D],
    ),
    (
        [Unordered.D, Unordered.C, Unordered.B, Unordered.A],
        [Unordered.D, Unordered.C, Unordered.B, Unordered.A],
    ),
    (
        [Unordered.C, Unordered.B, Unordered.A, Unordered.D],
        [Unordered.C, Unordered.B, Unordered.A, Unordered.D],
    ),
    (
        [Unordered.A, Unordered.A, Unordered.B, Unordered.B],
        [Unordered.A, Unordered.A, Unordered.B, Unordered.B],
    ),
    (
        [Unordered.C, Unordered.C, Unordered.B, Unordered.B],
        [Unordered.C, Unordered.C, Unordered.B, Unordered.B],
    ),
    # Ordered: should follow decorators order (A, B, C, D)
    ([Ordered.A, Ordered.B], [Ordered.A, Ordered.B]),
    ([Ordered.B, Ordered.A], [Ordered.A, Ordered.B]),
    ([Ordered.C, Ordered.D], [Ordered.C, Ordered.D]),
    ([Ordered.D, Ordered.C], [Ordered.C, Ordered.D]),
    (
        [Ordered.A, Ordered.B, Ordered.C, Ordered.D],
        [Ordered.A, Ordered.B, Ordered.C, Ordered.D],
    ),
    (
        [Ordered.D, Ordered.C, Ordered.B, Ordered.A],
        [Ordered.A, Ordered.B, Ordered.C, Ordered.D],
    ),
    (
        [Ordered.C, Ordered.B, Ordered.A, Ordered.D],
        [Ordered.A, Ordered.B, Ordered.C, Ordered.D],
    ),
    (
        [Ordered.A, Ordered.A, Ordered.B, Ordered.B],
        [Ordered.A, Ordered.A, Ordered.B, Ordered.B],
    ),
    (
        [Ordered.C, Ordered.C, Ordered.B, Ordered.B],
        [Ordered.B, Ordered.B, Ordered.C, Ordered.C],
    ),
]

METHOD_SEQUENCES = [
    # --------------------------------------------------------------
    # monitor class                 expected methods sequence
    # --------------------------------------------------------------
    (UnorderedMethodsMonitor, ["test_a", "test_b", "test_c"]),
    (OrderedMethodsMonitor, ["test_c", "test_b", "test_a"]),
    (EqualOrderedMethodsMonitor, ["test_a", "test_b", "test_c"]),
]


def test_suite_ordering():
    for monitors_sequence, expected_sequence in SUITE_SEQUENCES:
        suite = MonitorSuite()
        suite.add_monitors(monitors_sequence)
        sequence = [_extract_monitor_class(m) for m in suite]
        assert sequence == expected_sequence


def test_method_ordering():
    for monitor_class, expected_sequence in METHOD_SEQUENCES:
        suite = MonitorSuite()
        suite.add_monitor(monitor_class)
        sequence = [m.method_name for m in suite.all_monitors]
        assert sequence == expected_sequence


def _extract_monitor_class(suite):
    if suite.__class__ == MonitorSuite:
        return _extract_monitor_class(suite._tests[0])
    else:
        return suite.__class__



from spidermon import Monitor, MonitorSuite, monitors


# ----------------------------------
# Monitors
# ----------------------------------
class BaseMonitor(Monitor):
    def runTest(self):
        pass


class NotDescriptedMonitor(BaseMonitor):
    pass


class DocstringDescriptedMonitor(BaseMonitor):
    """docstring monitor description"""

    pass


@monitors.description("decorator monitor description")
class DescoratedDescriptedMonitor(BaseMonitor):
    pass


@monitors.description("decorator monitor description")
class DescoratedDescriptedMonitor2(BaseMonitor):
    """docstring monitor description"""


# ----------------------------------
# Suites
# ----------------------------------
class NotDescriptedSuite(MonitorSuite):
    pass


class DocstringDescriptedSuite(MonitorSuite):
    """docstring suite description"""

    pass


@monitors.description("decorator suite description")
class DescoratedDescriptedSuite(MonitorSuite):
    pass


@monitors.description("decorator suite description")
class DescoratedDescriptedSuite2(MonitorSuite):
    """docstring suite description"""


# ----------------------------------
# Methods
# ----------------------------------
class DescriptedMethodsMonitor(Monitor):
    def test_not_descripted(self):
        pass

    def test_docstring_descripted(self):
        """docstring method description"""
        pass

    @monitors.description("decorator method description")
    def test_decorator_descripted(self):
        pass

    @monitors.description("decorator method description")
    def test_decorator_descripted2(self):
        """docstring method description"""



from scrapy.settings import Settings

from spidermon.utils.settings import get_aws_credentials


def test_spidermon_aws_credentials_not_set():
    settings = Settings()

    (aws_access_key_id, aws_secret_access_key) = get_aws_credentials(settings)

    assert aws_access_key_id is None
    assert aws_secret_access_key is None


def test_spidermon_aws_credentials(mocker):
    warn_mock = mocker.patch("spidermon.utils.settings.warnings.warn")
    settings = Settings(
        {
            "SPIDERMON_AWS_ACCESS_KEY": "aws_access_key",
            "SPIDERMON_AWS_SECRET_KEY": "aws_secret_key",
        }
    )

    (aws_access_key_id, aws_secret_access_key) = get_aws_credentials(settings)

    assert aws_access_key_id == "aws_access_key"
    assert aws_secret_access_key == "aws_secret_key"
    warn_mock.assert_called_with(mocker.ANY, DeprecationWarning)


def test_spidermon_aws_credentials_scrapy_like():
    settings = Settings(
        {
            "SPIDERMON_AWS_ACCESS_KEY_ID": "aws_access_key_id",
            "SPIDERMON_AWS_SECRET_ACCESS_KEY": "aws_secret_access_key",
        }
    )

    (aws_access_key_id, aws_secret_access_key) = get_aws_credentials(settings)

    assert aws_access_key_id == "aws_access_key_id"
    assert aws_secret_access_key == "aws_secret_access_key"


def test_spidermon_aws_credentials_fall_back_to_scrapy():
    settings = Settings(
        {
            "AWS_ACCESS_KEY_ID": "scrapy_aws_access_key_id",
            "AWS_SECRET_ACCESS_KEY": "scrapy_aws_secret_access_key",
        }
    )

    (aws_access_key_id, aws_secret_access_key) = get_aws_credentials(settings)

    assert aws_access_key_id == "scrapy_aws_access_key_id"
    assert aws_secret_access_key == "scrapy_aws_secret_access_key"


def test_spidermon_aws_credentials_are_preferred_over_scrapy_ones():
    settings = Settings(
        {
            "AWS_ACCESS_KEY_ID": "scrapy_aws_access_key_id",
            "AWS_SECRET_ACCESS_KEY": "scrapy_aws_secret_access_key",
            "SPIDERMON_AWS_ACCESS_KEY_ID": "spidermon_aws_access_key_id",
            "SPIDERMON_AWS_SECRET_ACCESS_KEY": "spidermon_aws_secret_access_key",
        }
    )

    (aws_access_key_id, aws_secret_access_key) = get_aws_credentials(settings)

    assert aws_access_key_id == "spidermon_aws_access_key_id"
    assert aws_secret_access_key == "spidermon_aws_secret_access_key"


def test_spidermon_old_aws_credentials_are_preferred_over_new_ones(mocker):
    mocker.patch(
        "spidermon.utils.settings.warnings.warn"
    )  # avoid the warning in the tests
    settings = Settings(
        {
            "SPIDERMON_AWS_ACCESS_KEY": "old_aws_access_key",
            "SPIDERMON_AWS_SECRET_KEY": "old_aws_secret_key",
            "SPIDERMON_AWS_ACCESS_KEY_ID": "new_aws_access_key_id",
            "SPIDERMON_AWS_SECRET_ACCESS_KEY": "new_aws_secret_access_key",
        }
    )

    (aws_access_key_id, aws_secret_access_key) = get_aws_credentials(settings)

    assert aws_access_key_id == "old_aws_access_key"
    assert aws_secret_access_key == "old_aws_secret_key"



from unittest import mock
import pytest

from scrapy import signals
from spidermon.contrib.scrapy.extensions import Spidermon


@pytest.fixture
def suites():
    return ["tests.fixtures.suites.Suite01"]


def test_spider_opened_suites_should_run(get_crawler, suites):
    """The suites defined at spider_opened_suites should be loaded and run"""
    crawler = get_crawler()
    spidermon = Spidermon(crawler, spider_opened_suites=suites)
    spidermon.spider_opened_suites[0].run = mock.MagicMock()
    spidermon.spider_opened(crawler.spider)
    assert spidermon.spider_opened_suites[0].__class__.__name__ == "Suite01"
    spidermon.spider_opened_suites[0].run.assert_called_once_with(mock.ANY)


def test_spider_closed_suites_should_run(get_crawler, suites):
    """The suites defined at spider_closed_suites should be loaded and run"""
    crawler = get_crawler()
    spidermon = Spidermon(
        crawler, spider_opened_suites=suites, spider_closed_suites=suites
    )
    spidermon.spider_closed_suites[0].run = mock.MagicMock()
    spidermon.spider_opened(crawler.spider)
    spidermon.spider_closed(crawler.spider)
    assert spidermon.spider_closed_suites[0].__class__.__name__ == "Suite01"
    spidermon.spider_closed_suites[0].run.assert_called_once_with(mock.ANY)


def test_engine_stopped_suites_should_run(get_crawler, suites):
    """The suites defined at engine_stopped_suites should be loaded and run"""
    crawler = get_crawler()
    spidermon = Spidermon(crawler, engine_stopped_suites=suites)
    spidermon.engine_stopped_suites[0].run = mock.MagicMock()
    spidermon.engine_stopped()
    assert spidermon.engine_stopped_suites[0].__class__.__name__ == "Suite01"
    spidermon.engine_stopped_suites[0].run.assert_called_once_with(mock.ANY)


def test_spider_opened_suites_should_run_from_signal(get_crawler, suites):
    """The suites defined at SPIDERMON_SPIDER_OPEN_MONITORS setting should be loaded and run"""
    settings = {"SPIDERMON_SPIDER_OPEN_MONITORS": suites}
    crawler = get_crawler(settings)
    spidermon = Spidermon.from_crawler(crawler)
    spidermon.spider_opened_suites[0].run = mock.MagicMock()
    crawler.signals.send_catch_log(signal=signals.spider_opened, spider=crawler.spider)
    spidermon.spider_opened_suites[0].run.assert_called_once_with(mock.ANY)


def test_spider_closed_suites_should_run_from_signal(get_crawler, suites):
    """The suites defined at SPIDERMON_SPIDER_CLOSE_MONITORS setting should be loaded and run"""
    settings = {"SPIDERMON_SPIDER_CLOSE_MONITORS": suites}
    crawler = get_crawler(settings)
    spidermon = Spidermon.from_crawler(crawler)
    spidermon.spider_closed_suites[0].run = mock.MagicMock()
    crawler.signals.send_catch_log(signal=signals.spider_closed, spider=crawler.spider)
    spidermon.spider_closed_suites[0].run.assert_called_once_with(mock.ANY)


def test_engine_stopped_suites_should_run_from_signal(get_crawler, suites):
    """The suites defined at SPIDERMON_ENGINE_STOP_MONITORS setting should be loaded and run"""
    settings = {"SPIDERMON_ENGINE_STOP_MONITORS": suites}
    crawler = get_crawler(settings)
    spidermon = Spidermon.from_crawler(crawler)
    spidermon.engine_stopped_suites[0].run = mock.MagicMock()
    crawler.signals.send_catch_log(signal=signals.engine_stopped, spider=crawler.spider)
    spidermon.engine_stopped_suites[0].run.assert_called_once_with(mock.ANY)



import pytest

from .fixtures.names import *


NAMES = [
    "UnnamedMonitor/A Test",
    "UnnamedMonitor/test_without_name",
    "Class Monitor/A Test",
    "Class Monitor/test_without_name",
    "Instance Monitor/A Test",
    "Instance Monitor/test_without_name",
    "Instance Monitor/A Test",
    "Instance Monitor/test_without_name",
    "UnnamedMonitor/A Test",
    "UnnamedMonitor/test_without_name",
    "Class Monitor/A Test",
    "Class Monitor/test_without_name",
    "Instance Monitor/A Test",
    "Instance Monitor/test_without_name",
    "Instance Monitor/A Test",
    "Instance Monitor/test_without_name",
    "UnnamedMonitor/A Test",
    "UnnamedMonitor/test_without_name",
    "Class Monitor/A Test",
    "Class Monitor/test_without_name",
    "Instance Monitor/A Test",
    "Instance Monitor/test_without_name",
    "Instance Monitor/A Test",
    "Instance Monitor/test_without_name",
    "UnnamedMonitor/A Test",
    "UnnamedMonitor/test_without_name",
    "Class Monitor/A Test",
    "Class Monitor/test_without_name",
    "Instance Monitor/A Test",
    "Instance Monitor/test_without_name",
    "Instance Monitor/A Test",
    "Instance Monitor/test_without_name",
]

MONITOR_NAMES = [
    "UnnamedMonitor",
    "UnnamedMonitor",
    "Class Monitor",
    "Class Monitor",
    "Instance Monitor",
    "Instance Monitor",
    "Instance Monitor",
    "Instance Monitor",
    "UnnamedMonitor",
    "UnnamedMonitor",
    "Class Monitor",
    "Class Monitor",
    "Instance Monitor",
    "Instance Monitor",
    "Instance Monitor",
    "Instance Monitor",
    "UnnamedMonitor",
    "UnnamedMonitor",
    "Class Monitor",
    "Class Monitor",
    "Instance Monitor",
    "Instance Monitor",
    "Instance Monitor",
    "Instance Monitor",
    "UnnamedMonitor",
    "UnnamedMonitor",
    "Class Monitor",
    "Class Monitor",
    "Instance Monitor",
    "Instance Monitor",
    "Instance Monitor",
    "Instance Monitor",
]

METHOD_NAMES = [
    "A Test",
    "test_without_name",
    "A Test",
    "test_without_name",
    "A Test",
    "test_without_name",
    "A Test",
    "test_without_name",
    "A Test",
    "test_without_name",
    "A Test",
    "test_without_name",
    "A Test",
    "test_without_name",
    "A Test",
    "test_without_name",
    "A Test",
    "test_without_name",
    "A Test",
    "test_without_name",
    "A Test",
    "test_without_name",
    "A Test",
    "test_without_name",
    "A Test",
    "test_without_name",
    "A Test",
    "test_without_name",
    "A Test",
    "test_without_name",
    "A Test",
    "test_without_name",
]

FULL_NAMES_UNNNAMED_SUITE = [
    "UnnamedMonitor/A Test",
    "UnnamedMonitor/test_without_name",
    "Class Monitor/A Test",
    "Class Monitor/test_without_name",
    "Instance Monitor/A Test",
    "Instance Monitor/test_without_name",
    "Instance Monitor/A Test",
    "Instance Monitor/test_without_name",
    "The Child Suite/UnnamedMonitor/A Test",
    "The Child Suite/UnnamedMonitor/test_without_name",
    "The Child Suite/Class Monitor/A Test",
    "The Child Suite/Class Monitor/test_without_name",
    "The Child Suite/Instance Monitor/A Test",
    "The Child Suite/Instance Monitor/test_without_name",
    "The Child Suite/Instance Monitor/A Test",
    "The Child Suite/Instance Monitor/test_without_name",
    "Instance Suite Name/UnnamedMonitor/A Test",
    "Instance Suite Name/UnnamedMonitor/test_without_name",
    "Instance Suite Name/Class Monitor/A Test",
    "Instance Suite Name/Class Monitor/test_without_name",
    "Instance Suite Name/Instance Monitor/A Test",
    "Instance Suite Name/Instance Monitor/test_without_name",
    "Instance Suite Name/Instance Monitor/A Test",
    "Instance Suite Name/Instance Monitor/test_without_name",
    "Instance Suite Name/UnnamedMonitor/A Test",
    "Instance Suite Name/UnnamedMonitor/test_without_name",
    "Instance Suite Name/Class Monitor/A Test",
    "Instance Suite Name/Class Monitor/test_without_name",
    "Instance Suite Name/Instance Monitor/A Test",
    "Instance Suite Name/Instance Monitor/test_without_name",
    "Instance Suite Name/Instance Monitor/A Test",
    "Instance Suite Name/Instance Monitor/test_without_name",
]

FULL_NAMES_NAMED_SUITE = [
    "The Top Suite/UnnamedMonitor/A Test",
    "The Top Suite/UnnamedMonitor/test_without_name",
    "The Top Suite/Class Monitor/A Test",
    "The Top Suite/Class Monitor/test_without_name",
    "The Top Suite/Instance Monitor/A Test",
    "The Top Suite/Instance Monitor/test_without_name",
    "The Top Suite/Instance Monitor/A Test",
    "The Top Suite/Instance Monitor/test_without_name",
    "The Top Suite/The Child Suite/UnnamedMonitor/A Test",
    "The Top Suite/The Child Suite/UnnamedMonitor/test_without_name",
    "The Top Suite/The Child Suite/Class Monitor/A Test",
    "The Top Suite/The Child Suite/Class Monitor/test_without_name",
    "The Top Suite/The Child Suite/Instance Monitor/A Test",
    "The Top Suite/The Child Suite/Instance Monitor/test_without_name",
    "The Top Suite/The Child Suite/Instance Monitor/A Test",
    "The Top Suite/The Child Suite/Instance Monitor/test_without_name",
    "The Top Suite/Instance Suite Name/UnnamedMonitor/A Test",
    "The Top Suite/Instance Suite Name/UnnamedMonitor/test_without_name",
    "The Top Suite/Instance Suite Name/Class Monitor/A Test",
    "The Top Suite/Instance Suite Name/Class Monitor/test_without_name",
    "The Top Suite/Instance Suite Name/Instance Monitor/A Test",
    "The Top Suite/Instance Suite Name/Instance Monitor/test_without_name",
    "The Top Suite/Instance Suite Name/Instance Monitor/A Test",
    "The Top Suite/Instance Suite Name/Instance Monitor/test_without_name",
    "The Top Suite/Instance Suite Name/UnnamedMonitor/A Test",
    "The Top Suite/Instance Suite Name/UnnamedMonitor/test_without_name",
    "The Top Suite/Instance Suite Name/Class Monitor/A Test",
    "The Top Suite/Instance Suite Name/Class Monitor/test_without_name",
    "The Top Suite/Instance Suite Name/Instance Monitor/A Test",
    "The Top Suite/Instance Suite Name/Instance Monitor/test_without_name",
    "The Top Suite/Instance Suite Name/Instance Monitor/A Test",
    "The Top Suite/Instance Suite Name/Instance Monitor/test_without_name",
]


@pytest.fixture
def named_top_suite():
    return NamedTopSuite()


@pytest.fixture
def unnamed_top_suite():
    return UnnamedTopSuite()


def test_names(named_top_suite, unnamed_top_suite):
    _check_names(generated_names=_generate_names(named_top_suite), expected_names=NAMES)
    _check_names(
        generated_names=_generate_names(unnamed_top_suite), expected_names=NAMES
    )


def test_monitor_names(named_top_suite, unnamed_top_suite):
    _check_names(
        generated_names=_generate_monitor_names(named_top_suite),
        expected_names=MONITOR_NAMES,
    )
    _check_names(
        generated_names=_generate_monitor_names(unnamed_top_suite),
        expected_names=MONITOR_NAMES,
    )


def test_method_names(named_top_suite, unnamed_top_suite):
    _check_names(
        generated_names=_generate_method_names(named_top_suite),
        expected_names=METHOD_NAMES,
    )
    _check_names(
        generated_names=_generate_method_names(unnamed_top_suite),
        expected_names=METHOD_NAMES,
    )


def test_full_names(named_top_suite, unnamed_top_suite):
    _check_names(
        generated_names=_generate_full_names(named_top_suite),
        expected_names=FULL_NAMES_NAMED_SUITE,
    )
    _check_names(
        generated_names=_generate_full_names(unnamed_top_suite),
        expected_names=FULL_NAMES_UNNNAMED_SUITE,
    )


def _generate_names(suite):
    return [test.name for test in suite.all_monitors]


def _generate_monitor_names(suite):
    return [test.monitor_name for test in suite.all_monitors]


def _generate_method_names(suite):
    return [test.method_name for test in suite.all_monitors]


def _generate_full_names(suite):
    return [test.full_name for test in suite.all_monitors]


def _check_names(generated_names, expected_names):
    assert len(generated_names) == len(expected_names)
    for generated_name, expected_name in zip(generated_names, expected_names):
        assert generated_name == expected_name



import logging

import json
from scrapy.utils.misc import load_object
from urllib.parse import urlparse
from urllib.request import urlopen

logger = logging.getLogger(__name__)


def get_schema_from(source):
    if is_schema_url(source):
        schema = get_contents(source)
        try:
            return json.loads(schema)
        except Exception as e:
            logger.exception(str(e) + f"\nCould not parse schema from '{source}'")
    elif source.endswith(".json"):
        with open(source) as f:
            try:
                return json.load(f)
            except Exception as e:
                logger.exception(str(e) + f"\nCould not parse schema in '{source}'")
    else:
        schema = load_object(source)
        if isinstance(schema, str):
            return json.loads(schema)
        return schema


def is_schema_url(path):
    result = urlparse(path)
    try:
        return all([result.scheme, result.netloc, result.path])
    except AttributeError:
        return False


def get_contents(url):
    try:
        with urlopen(url) as f:
            return f.read().decode("utf-8")
    except Exception as e:
        logger.exception(str(e) + f"\nFailed to get '{url}'")



from spidermon import Monitor, MonitorSuite, monitors


class DummyMonitor(Monitor):
    def runTest(self):
        pass


class DummyMonitorSuite(MonitorSuite):
    monitors = [DummyMonitor]


# ----------------------------------
# Monitors ordering
# ----------------------------------
class Unordered:
    class A(DummyMonitor):
        pass

    class B(DummyMonitor):
        pass

    class C(DummyMonitorSuite):
        pass

    class D(DummyMonitorSuite):
        pass


class Ordered:
    @monitors.order(1)
    class A(DummyMonitor):
        pass

    @monitors.order(2)
    class B(DummyMonitor):
        pass

    @monitors.order(3)
    class C(DummyMonitorSuite):
        pass

    @monitors.order(4)
    class D(DummyMonitorSuite):
        pass


# ----------------------------------
# Methods ordering
# ----------------------------------
class UnorderedMethodsMonitor(Monitor):
    def test_a(self):
        pass

    def test_b(self):
        pass

    def test_c(self):
        pass


class OrderedMethodsMonitor(Monitor):
    @monitors.order(3)
    def test_a(self):
        pass

    @monitors.order(2)
    def test_b(self):
        pass

    @monitors.order(1)
    def test_c(self):
        pass


class EqualOrderedMethodsMonitor(Monitor):
    @monitors.order(5)
    def test_a(self):
        pass

    @monitors.order(5)
    def test_b(self):
        pass

    @monitors.order(5)
    def test_c(self):
        pass



import sys

import pytest
from scrapy import signals
from scrapy.spiders import Spider
from scrapy.utils.test import get_crawler

from spidermon.contrib.scrapy.extensions import Spidermon


@pytest.fixture
def spidermon_enabled_settings():
    return {
        "SPIDERMON_ENABLED": True,
        "EXTENSIONS": {"spidermon.contrib.scrapy.extensions.Spidermon": 100},
    }


def test_spider_opened_connect_signal(mocker, spidermon_enabled_settings):
    spider_opened_method = mocker.patch.object(Spidermon, "spider_opened")

    crawler = get_crawler(settings_dict=spidermon_enabled_settings)
    spider = Spider.from_crawler(crawler, "example.com")
    crawler.signals.send_catch_log(signal=signals.spider_opened, spider=spider)

    assert spider_opened_method.called, "spider_opened not called"


def test_spider_closed_connect_signal(mocker, spidermon_enabled_settings):
    spider_closed_method = mocker.patch.object(Spidermon, "spider_closed")

    crawler = get_crawler(settings_dict=spidermon_enabled_settings)
    spider = Spider.from_crawler(crawler, "example.com")
    crawler.signals.send_catch_log(
        signal=signals.spider_closed, spider=spider, reason=None
    )

    assert spider_closed_method.called, "spider_closed not called"


def test_engine_stopped_connect_signal(mocker, spidermon_enabled_settings):
    engine_stopped = mocker.patch.object(Spidermon, "engine_stopped")

    crawler = get_crawler(settings_dict=spidermon_enabled_settings)
    spider = Spider.from_crawler(crawler, "example.com")
    crawler.signals.send_catch_log(
        signal=signals.engine_stopped, spider=spider, reason=None
    )

    assert engine_stopped.called, "engine_stopped not called"


def test_item_scraped_connect_signal_if_field_coverage_settings_enabled(
    mocker, spidermon_enabled_settings
):
    item_scraped_method = mocker.patch.object(Spidermon, "item_scraped")

    spidermon_enabled_settings["SPIDERMON_ADD_FIELD_COVERAGE"] = True
    crawler = get_crawler(settings_dict=spidermon_enabled_settings)

    spider = Spider.from_crawler(crawler, "example.com")
    crawler.signals.send_catch_log(signal=signals.item_scraped, spider=spider)

    assert item_scraped_method.called, "item_scraped_method not called"


def test_item_scraped_do_not_connect_signal_if_field_coverage_settings_disabled(
    mocker, spidermon_enabled_settings
):
    item_scraped_method = mocker.patch.object(Spidermon, "item_scraped")

    spidermon_enabled_settings["SPIDERMON_ADD_FIELD_COVERAGE"] = False

    crawler = get_crawler(settings_dict=spidermon_enabled_settings)

    spider = Spider.from_crawler(crawler, "example.com")
    crawler.signals.send_catch_log(signal=signals.item_scraped, spider=spider)

    assert not item_scraped_method.called, "item_scraped_method called"


def test_item_scraped_do_not_connect_signal_if_do_not_have_field_coverage_settings(
    mocker, spidermon_enabled_settings
):
    item_scraped_method = mocker.patch.object(Spidermon, "item_scraped")

    crawler = get_crawler(settings_dict=spidermon_enabled_settings)

    spider = Spider.from_crawler(crawler, "example.com")
    crawler.signals.send_catch_log(signal=signals.item_scraped, spider=spider)

    assert not item_scraped_method.called, "item_scraped_method called"



import inspect

from spidermon.exceptions import (
    InvalidMonitor,
    InvalidMonitorClass,
    InvalidMonitorTuple,
)
from .monitors import Monitor
from .actions import Action


class MonitorFactory:
    @classmethod
    def load_monitor(cls, monitor, name=None):
        from .suites import MonitorSuite

        if inspect.isclass(monitor):
            return cls.load_monitor_from_class(monitor_class=monitor, name=name)
        elif isinstance(monitor, tuple):
            return cls.load_monitor_from_tuple(monitor_tuple=monitor)
        elif isinstance(monitor, (Monitor, MonitorSuite)):
            return monitor
        cls.raise_invalid_monitor()

    @classmethod
    def load_monitor_from_class(cls, monitor_class, name=None):
        from .suites import MonitorSuite

        if issubclass(monitor_class, Monitor):
            from spidermon.loaders import MonitorLoader

            loader = MonitorLoader()
            monitor = loader.load_suite_from_monitor(
                monitor_class=monitor_class, name=name
            )
        elif issubclass(monitor_class, MonitorSuite):
            monitor = monitor_class(name=name)
        else:
            cls.raise_invalid_class()
        return cls.load_monitor(monitor=monitor, name=name)

    @classmethod
    def load_monitor_from_tuple(cls, monitor_tuple):
        if len(monitor_tuple) != 2:
            cls.raise_invalid_tuple()
        name, monitor = monitor_tuple
        if not isinstance(name, str):
            cls.raise_invalid_tuple()
        return cls.load_monitor(monitor=monitor, name=name)

    @classmethod
    def raise_invalid_monitor(cls):
        raise InvalidMonitor(
            "Wrong Monitor definition, it should be:\n"
            "- an instance of a Monitor/MonitorSuite object.\n"
            "- a subclass of Monitor/MonitorSuite.\n"
            "- a tuple with the format (name, monitor).\n"
            "- a string containing an evaluable python expression."
        )

    @classmethod
    def raise_invalid_class(cls):
        raise InvalidMonitorClass(
            "Wrong Monitor class definition, it should be "
            "an instance of a Monitor/MonitorSuite object."
        )

    @classmethod
    def raise_invalid_tuple(cls):
        raise InvalidMonitorTuple(
            "Wrong Monitor tuple definition, it should be "
            "a tuple with the format (name, monitor)"
        )


class ActionFactory:
    @classmethod
    def load_action(cls, action, crawler=None):
        if inspect.isclass(action):
            return cls.load_action_from_class(action_class=action, crawler=crawler)
        elif isinstance(action, Action):
            return action
        cls.raise_invalid_action()

    @classmethod
    def load_action_from_class(cls, action_class, crawler=None):
        if not issubclass(action_class, Action):
            cls.raise_invalid_class()
        if crawler and hasattr(action_class, "from_crawler"):
            return action_class.from_crawler(crawler)
        else:
            return action_class()

    @classmethod
    def raise_invalid_action(cls):
        raise Exception
        raise InvalidMonitor(
            "Wrong Monitor definition, it should be:\n"
            "- an instance of a Monitor/MonitorSuite object.\n"
            "- a subclass of Monitor/MonitorSuite.\n"
            "- a tuple with the format (name, monitor).\n"
            "- a string containing an evaluable python expression."
        )

    @classmethod
    def raise_invalid_class(cls):
        raise Exception
        raise InvalidMonitorClass(
            "Wrong Monitor class definition, it should be "
            "an instance of a Monitor/MonitorSuite object."
        )



import warnings

from spidermon.contrib.stats.counters import (
    PercentCounter,
    DictPercentCounter,
    AttributeDictPercentCounter,
)
from spidermon.contrib.stats.analyzer import StatsAnalyzer
from spidermon.contrib.scrapy.stats import STATS_DEFAULT_VALIDATION_PREFIX

from .stats import StatsMonitorMixin


class MetaDictPercentCounter(DictPercentCounter):
    def add_values(self, key, subkey, value):
        if key not in self._dict:
            self._create_item(key)
        self[key].add_value(subkey, value)


class ErrorsDictPercentCounter(AttributeDictPercentCounter):
    __attribute_dict_name__ = "fields"


class ErrorsInfo(MetaDictPercentCounter):
    __items_class__ = ErrorsDictPercentCounter


class FieldErrorsDictPercentCounter(AttributeDictPercentCounter):
    __attribute_dict_name__ = "errors"


class FieldErrorsInfo(MetaDictPercentCounter):
    __items_class__ = FieldErrorsDictPercentCounter

    def __init__(self, fields_count, items_count):
        super().__init__(items_count)
        self._fields_count = fields_count

    @property
    def count(self):
        return self._fields_count


class ItemsInfo:
    def __init__(self, items_count, items_with_errors, items_dropped):
        self.count = items_count
        self.errors = PercentCounter(count=items_with_errors, total=items_count)
        self.dropped = PercentCounter(count=items_dropped, total=items_count)


class ValidationInfo:
    def __init__(self, stats, prefix=None):
        self.analyzer = StatsAnalyzer(
            stats=stats, prefix=prefix or STATS_DEFAULT_VALIDATION_PREFIX
        )

        # items
        items_count = sum(self.analyzer.search("items$").values())
        items_with_errors_count = sum(self.analyzer.search("items/errors$").values())
        items_dropped_count = sum(self.analyzer.search("items/dropped$").values())
        self.items = ItemsInfo(
            items_count=items_count,
            items_with_errors=items_with_errors_count,
            items_dropped=items_dropped_count,
        )

        # errors & fields
        fields_count = sum(self.analyzer.search("fields$").values())
        self.errors = ErrorsInfo(items_count)
        self.fields = FieldErrorsInfo(
            fields_count=fields_count, items_count=items_count
        )

        field_errors = self.analyzer.search(
            "fields/errors/([^/]+)$", include_matches=True
        )
        for _, error in field_errors.values():
            field_errors_per_field = self.analyzer.search(
                "fields/errors/%s/([^/]+)$" % error, include_matches=True
            )
            for count, field in field_errors_per_field.values():
                self.errors.add_values(key=error, subkey=field, value=count)
                self.fields.add_values(key=field, subkey=error, value=count)


class ValidationMonitorMixin(StatsMonitorMixin):

    correct_field_list_handling = False

    @property
    def validation(self):
        if not hasattr(self, "_validation"):
            self._validation = ValidationInfo(self.stats)
        return self._validation

    def _get_all_fields(self):
        return sorted(self.validation.fields)

    @staticmethod
    def _warn_list_handling():
        warnings.warn(
            "ValidationMonitorMixin behavior with is deprecated, please switch to per-field thresholds and set the "
            '"correct_field_list_handling" class attribute',
            DeprecationWarning,
        )

    def check_missing_required_fields(self, field_names=None, allowed_count=0):
        """
        Checks that the number of "missing_required_field" errors for the ``field_names`` fields is less or equal than
        ``allowed_count`` and raises an error with all problematic fields.
        If ``field_names`` is None and ``self.correct_field_list_handling`` is True, checks all fields.
        If ``field_names`` is None and ``self.correct_field_list_handling`` is False, checks that the total number of
        "missing_required_field" errors is less or equal than ``allowed_count``.
        """
        if not self.correct_field_list_handling and not field_names:
            self._warn_list_handling()
            missing_count = self.validation.errors["missing_required_field"].count
            self.assertLessEqual(
                missing_count,
                allowed_count,
                msg="{count} required field{plural} are missing!{threshold_info}".format(
                    count=missing_count,
                    plural="" if missing_count == 1 else "s",
                    threshold_info=(" (maximum allowed %d)" % allowed_count)
                    if allowed_count > 0
                    else "",
                ),
            )
            return
        if not field_names:
            field_names = self._get_all_fields()
        msgs = []
        for field_name in field_names:
            missing_count = (
                self.validation.fields[field_name]
                .errors["missing_required_field"]
                .count
            )
            if missing_count > allowed_count:
                msg = self._get_msg_for_missing_required_count(
                    field_name, missing_count, allowed_count
                )
                msgs.append(msg)
        if msgs:
            msgs.insert(0, "Required fields are missing:")
            self.fail("\n".join(msgs))

    def check_missing_required_field(self, field_name, allowed_count=0):
        """
        Checks that the number of "missing_required_field" errors for the ``field_name`` field is less or equal than
        ``allowed_count``.
        """
        missing_count = (
            self.validation.fields[field_name].errors["missing_required_field"].count
        )
        msg = self._get_msg_for_missing_required_count(
            field_name, missing_count, allowed_count
        )
        self.assertLessEqual(missing_count, allowed_count, msg)

    @staticmethod
    def _get_msg_for_missing_required_count(field_name, missing_count, allowed_count):
        msg = "Required field {field} is missing in {count} items!{threshold_info}".format(
            count=missing_count,
            field=field_name,
            threshold_info=(" (maximum allowed %d)" % allowed_count)
            if allowed_count > 0
            else "",
        )
        return msg

    def check_missing_required_fields_percent(
        self, field_names=None, allowed_percent=0
    ):
        """
        Checks that the number of "missing_required_field" errors for the ``field_names`` fields divided by the number
        of items is less or equal than ``allowed_percent`` and raises an error with all problematic fields.
        If ``field_names`` is None and ``self.correct_field_list_handling`` is True, checks all fields.
        If ``field_names`` is None and ``self.correct_field_list_handling`` is False, checks that the total number of
        "missing_required_field" errors is less or equal than ``allowed_count``.
        """
        if not self.correct_field_list_handling and not field_names:
            self._warn_list_handling()
            missing_percent = self.validation.errors["missing_required_field"].percent
            self.assertLessEqual(
                missing_percent,
                allowed_percent,
                msg="{percent}% of required fields are missing!{threshold_info}".format(
                    percent=missing_percent * 100,
                    threshold_info=(
                        " (maximum allowed %.0f%%)" % (allowed_percent * 100)
                    )
                    if allowed_percent > 0
                    else "",
                ),
            )
            return
        if not field_names:
            field_names = self._get_all_fields()
        msgs = []
        for field_name in field_names:
            missing_percent = (
                self.validation.fields[field_name]
                .errors["missing_required_field"]
                .percent
            )
            if missing_percent > allowed_percent:
                msg = self._get_msg_for_missing_required_percent(
                    field_name, missing_percent, allowed_percent
                )
                msgs.append(msg)
        if msgs:
            msgs.insert(0, "Required fields are missing:")
            self.fail("\n".join(msgs))

    def check_missing_required_field_percent(self, field_name, allowed_percent=0):
        """
        Checks that the number of "missing_required_field" errors for the ``field_name`` field divided by the number
        of items is less or equal than ``allowed_percent``.
        """
        missing_percent = (
            self.validation.fields[field_name].errors["missing_required_field"].percent
        )
        msg = self._get_msg_for_missing_required_percent(
            field_name, missing_percent, allowed_percent
        )
        self.assertLessEqual(missing_percent, allowed_percent, msg)

    @staticmethod
    def _get_msg_for_missing_required_percent(
        field_name, missing_percent, allowed_percent
    ):
        msg = (
            "{percent}% of required field {field} are missing!{threshold_info}".format(
                percent=missing_percent * 100,
                field=field_name,
                threshold_info=(" (maximum allowed %.0f%%)" % (allowed_percent * 100))
                if allowed_percent > 0
                else "",
            )
        )
        return msg

    def check_fields_errors(self, field_names=None, errors=None, allowed_count=0):
        """
        Checks that the number of errors for the ``field_names`` fields is less or equal than ``allowed_count`` and
        raises an error with all problematic fields.
        If ``field_names`` is None and ``self.correct_field_list_handling`` is True, checks all fields.
        If ``field_names`` is None and ``self.correct_field_list_handling`` is False, checks that the total number of errors is less or
        equal than ``allowed_count``.
        """
        if not self.correct_field_list_handling and not field_names:
            self._warn_list_handling()
            errors_count = self.validation.errors.count
            self.assertLessEqual(
                errors_count,
                allowed_count,
                msg="{count} field{count_plural} {verb} validation errors!{threshold_info}".format(
                    count=errors_count,
                    count_plural="" if errors_count == 1 else "s",
                    verb="has" if errors_count == 1 else "have",
                    threshold_info=(" (maximum allowed %d)" % allowed_count)
                    if allowed_count > 0
                    else "",
                ),
            )
            return
        if not field_names:
            field_names = self._get_all_fields()
        msgs = []
        for field_name in field_names:
            errors_count = self._get_errors_count(errors, field_name)
            if errors_count > allowed_count:
                msg = self._get_msg_for_field_errors(
                    field_name, errors_count, allowed_count
                )
                msgs.append(msg)
        if msgs:
            msgs.insert(0, "There are field errors:")
            self.fail("\n".join(msgs))

    def check_field_errors(self, field_name, errors=None, allowed_count=0):
        """
        Checks that the number of errors for the ``field_name`` field is less or equal than ``allowed_count``.
        """
        errors_count = self._get_errors_count(errors, field_name)
        msg = self._get_msg_for_field_errors(field_name, errors_count, allowed_count)
        self.assertLessEqual(errors_count, allowed_count, msg)

    def _get_errors_count(self, errors, field_name):
        if errors:
            errors_count = sum(
                [self.validation.fields[field_name].errors[e].count for e in errors]
            )
        else:
            errors_count = self.validation.fields[field_name].errors.count
        return errors_count

    @staticmethod
    def _get_msg_for_field_errors(field_name, errors_count, allowed_count):
        msg = "Field {field} has {count} validation errors!{threshold_info}".format(
            count=errors_count,
            field=field_name,
            threshold_info=(" (maximum allowed %d)" % allowed_count)
            if allowed_count > 0
            else "",
        )
        return msg

    def check_fields_errors_percent(
        self, field_names=None, errors=None, allowed_percent=0
    ):
        """
        Checks that the number of errors for the ``field_names`` fields divided by the number of items is less or equal
        than ``allowed_percent`` and raises an error with all problematic fields.
        If ``field_names`` is None and ``self.correct_field_list_handling`` is True, checks all fields.
        If ``field_names`` is None and ``self.correct_field_list_handling`` is False, checks that the total number of errors divided by the
        number of items is less or equal than ``allowed_count``
        """
        if not self.correct_field_list_handling and not field_names:
            self._warn_list_handling()
            errors_percent = self.validation.errors.percent
            self.assertLessEqual(
                errors_percent,
                allowed_percent,
                msg="{percent}% of fields have validation errors!{threshold_info}".format(
                    percent=errors_percent * 100,
                    threshold_info=(
                        " (maximum allowed %.0f%%)" % (allowed_percent * 100)
                    )
                    if allowed_percent > 0
                    else "",
                ),
            )
            return
        if not field_names:
            field_names = self._get_all_fields()
        msgs = []
        for field_name in field_names:
            errors_percent = self._get_errors_percent(errors, field_name)
            if errors_percent > allowed_percent:
                msg = self._get_msg_for_field_errors_percent(
                    field_name, errors_percent, allowed_percent
                )
                msgs.append(msg)
        if msgs:
            msgs.insert(0, "There are field errors:")
            self.fail("\n".join(msgs))

    def check_field_errors_percent(self, field_name, errors=None, allowed_percent=0):
        """
        Checks that the number of errors for the ``field_name`` field divided by the number of items is less or equal
        than ``allowed_percent``.
        """
        errors_percent = self._get_errors_percent(errors, field_name)
        msg = self._get_msg_for_field_errors_percent(
            field_name, errors_percent, allowed_percent
        )
        self.assertLessEqual(errors_percent, allowed_percent, msg)

    def _get_errors_percent(self, errors, field_name):
        if errors:
            errors_percent = sum(
                [self.validation.fields[field_name].errors[e].percent for e in errors]
            )
        else:
            errors_percent = self.validation.fields[field_name].errors.percent
        return errors_percent

    @staticmethod
    def _get_msg_for_field_errors_percent(field_name, errors_percent, allowed_percent):
        msg = "{percent}% of field {field} have validation errors!{threshold_info}".format(
            percent=errors_percent * 100,
            field=field_name,
            threshold_info=(" (maximum allowed %.0f%%)" % (allowed_percent * 100))
            if allowed_percent > 0
            else "",
        )
        return msg



from spidermon import Monitor


class EmptyMonitor(Monitor):
    pass


class Monitor01(Monitor):
    def test_a(self):
        pass

    def test_b(self):
        pass

    def test_c(self):
        pass


class Monitor02(Monitor):
    def test_d(self):
        pass

    def test_e(self):
        pass



from setuptools import setup, find_packages

test_requirements = [
    "pytest>=2.7.0",
    "pytest-cov",
    "pytest-mock",
    "jinja2",
    "lxml",
    "premailer",
    "scrapy",
    "slackclient",
    "twisted>=19.7.0",
]

setup(
    name="spidermon",
    version="1.16.2",
    url="https://github.com/scrapinghub/spidermon",
    author="Zyte",
    author_email="opensource@zyte.com",
    description=("Spidermon is a framework to build monitors for Scrapy spiders."),
    long_description=("Spidermon is a framework to build monitors for Scrapy spiders."),
    license="BSD",
    packages=find_packages(),
    package_data={"spidermon": ["VERSION"]},
    zip_safe=False,
    include_package_data=True,
    install_requires=["jsonschema[format]==3.2.0", "python-slugify"],
    tests_require=test_requirements,
    extras_require={
        # Specific monitors and tools to support notifications and reports
        "monitoring": [
            "scrapy",
            "Jinja2",
            "slackclient",
            "boto",
            "boto3",
            "premailer",
            "sentry-sdk",
        ],
        # Data validation
        "validation": ["schematics"],
        # Tools to run the tests
        "tests": test_requirements,
        "pep8": ["black"],
        # Tools to build and publish the documentation
        "docs": ["sphinx", "sphinx-rtd-theme", "s3cmd"],
    },
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Framework :: Scrapy",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: System :: Monitoring",
    ],
    python_requires=">=3.6",
)



import sys

from spidermon.core.suites import MonitorSuite
from spidermon.results.monitor import MonitorResult
from spidermon.results.text import TextMonitorResult
from spidermon.exceptions import InvalidMonitor, InvalidResult
from spidermon.data import Data


class MonitorRunner:
    data_immutable_dicts = ["stats"]
    data_default_data = {"meta": {}}

    def __init__(self):
        self.suite = None
        self.result = None
        self.data = None

    def run(self, suite, **data):
        if not isinstance(suite, MonitorSuite):
            raise InvalidMonitor("Runners must receive a MonitorSuite instance")
        self.suite = suite
        data = dict(self.data_default_data, **data)
        self.data = self.transform_data(**data)
        self.suite.init_data(self.data)
        self.result = self.create_result()
        if not isinstance(self.result, MonitorResult):
            raise InvalidResult("Runners must use a MonitorResult instance")
        return self.run_suite()

    def transform_data(self, **data):
        data = data or {}
        new_data_dict = {}
        for attr_name, attr in data.items():
            if attr_name in self.data_immutable_dicts:
                new_data = Data(attr)
            else:
                new_data = attr
            new_data_dict[attr_name] = new_data
        return Data(new_data_dict)

    def run_suite(self):
        self.result.start()
        self.run_monitors()
        self.run_actions()
        self.result.finish()
        return self.result

    def run_monitors(self):
        self.result.next_step()
        self.suite(self.result)
        self.result.finish_step()

    def run_actions(self):
        # Run monitors finished actions
        self.result.next_step()
        self.run_monitors_finished()
        self.result.finish_step()

        # Run monitors passed actions
        self.result.next_step()
        if self.result.monitor_results and self.result.all_monitors_passed:
            self.run_monitors_passed()
        else:
            self.result.skip_all_step_actions(
                actions=self.suite.monitors_passed_actions, reason="A Monitor failed"
            )
        self.result.finish_step()

        # Run monitors failed actions
        self.result.next_step()
        if self.result.monitor_results and not self.result.all_monitors_passed:
            self.run_monitors_failed()
        else:
            self.result.skip_all_step_actions(
                actions=self.suite.monitors_failed_actions, reason="No Monitors failed"
            )
        self.result.finish_step()

    def run_monitors_finished(self):
        self.suite.on_monitors_finished(self.result)
        for action in self.suite.monitors_finished_actions:
            action.run(self.result, self.data)

    def run_monitors_passed(self):
        self.suite.on_monitors_passed(self.result)
        for action in self.suite.monitors_passed_actions:
            action.run(self.result, self.data)

    def run_monitors_failed(self):
        self.suite.on_monitors_failed(self.result)
        for action in self.suite.monitors_failed_actions:
            action.run(self.result, self.data)

    def create_result(self):
        return MonitorResult()


class TextMonitorRunner(MonitorRunner):
    def __init__(self, stream=sys.stderr, verbosity=1):
        super().__init__()
        self.stream = stream
        self.verbosity = verbosity

    def create_result(self):
        return TextMonitorResult(stream=self.stream, verbosity=self.verbosity)



from spidermon.exceptions import NotConfigured


class Context(dict):
    """
    Stores context for python expressions.

    Also keeps track of not configured components (variables) of the context
    to throw NotConfigured exception in the right time, when test is
    evaluated by interpreter, instead of throwing it in building-context time
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._notconfigured = []

    def __getitem__(self, item):
        if item in self._notconfigured:
            raise NotConfigured(f"{item} not available!")
        return super().__getitem__(item)

    def extend_via_attrs(self, obj, attrs):
        """Extend context with names of object attributes and their values"""
        for attr in attrs:
            try:
                super().__setitem__(attr, getattr(obj, attr))
            except NotConfigured:
                self._notconfigured.append(attr)



import os

import jinja2
import pytest
from spidermon.templates import TemplateLoader


def create_file(path, content):
    with open(path, "w") as file:
        file.write(content)


def test_get_template_with_absolute_path(tmpdir):
    template_loader = TemplateLoader()
    test_template = os.path.join(str(tmpdir), "absolute_path_template.jinja")
    create_file(test_template, "test content")

    assert tmpdir not in template_loader.paths
    template = template_loader.get_template(test_template)
    assert template.render() == "test content"


def test_fail_get_absolute_template_that_does_not_exist(tmpdir):
    template_loader = TemplateLoader()
    test_template = os.path.join(str(tmpdir), "i_do_not_exist.template")
    with pytest.raises(jinja2.exceptions.TemplateNotFound):
        template = template_loader.get_template(test_template)  # noqa


def test_fail_get_template_that_does_not_exist(tmpdir):
    template_loader = TemplateLoader()
    with pytest.raises(jinja2.exceptions.TemplateNotFound):
        template = template_loader.get_template("i_do_not_exist.template")  # noqa



from unittest import TestCase

from spidermon.contrib.validation import JSONSchemaValidator
from spidermon.contrib.validation import messages

from slugify import slugify


class SchemaTestCaseMetaclass(type):
    def __new__(mcs, name, bases, attrs):
        def _test_function(data_test):
            def _function(self):
                validator = JSONSchemaValidator(data_test.schema or self.schema)
                assert validator.validate(data_test.data) == (
                    data_test.valid,
                    data_test.expected_errors,
                )

            return _function

        cls = super().__new__(mcs, name, bases, attrs)
        for dt in getattr(cls, "data_tests", []):
            function_name = "test_%s" % slugify(dt.name, separator="_").lower()
            setattr(cls, function_name, _test_function(dt))
        return cls


class SchemaTest(TestCase, metaclass=SchemaTestCaseMetaclass):
    schema = {}
    data_tests = []


class DataTest:
    def __init__(self, name, data, valid, expected_errors=None, schema=None):
        self.name = name
        self.data = data
        self.valid = valid
        self.expected_errors = expected_errors or {}
        self.schema = schema


class AdditionalItems(SchemaTest):
    data_tests = [
        DataTest(
            name="additionalItems as schema, additional items match schema",
            schema={"items": [{}], "additionalItems": {"type": "integer"}},
            data=[None, 2, 3, 4],
            valid=True,
        ),
        DataTest(
            name="additionalItems as schema, additional items do not match schema",
            schema={"items": [{}], "additionalItems": {"type": "integer"}},
            data=[None, 2, 3, "foo"],
            valid=False,
            expected_errors={"3": [messages.INVALID_INT]},
        ),
        DataTest(
            name="no additionalItems, all items match schema",
            schema={"items": {}, "additionalItems": False},
            data=[1, 2, 3, 4, 5],
            valid=True,
        ),
        DataTest(
            name="array of items with no additionalItems, no additional items present",
            schema={"items": [{}, {}, {}], "additionalItems": False},
            data=[1, 2, 3],
            valid=True,
        ),
        DataTest(
            name="array of items with no additionalItems, additional items are not permitted",
            schema={"items": [{}, {}, {}], "additionalItems": False},
            data=[1, 2, 3, 4],
            valid=False,
            expected_errors={"": [messages.TOO_MANY_ITEMS]},
        ),
        DataTest(
            name="additionalItems as false without items, valid",
            schema={"additionalItems": False},
            data=[1, 2, 3, 4, 5],
            valid=True,
        ),
        DataTest(
            name="additionalItems as false without items, ignores non-arrays",
            schema={"additionalItems": False},
            data={"foo": "bar"},
            valid=True,
        ),
        DataTest(
            name="additionalItems are allowed by default, only the first item is validated",
            schema={"items": [{"type": "integer"}]},
            data=[1, "foo", False],
            valid=True,
        ),
        DataTest(
            name="array, sequence without extra item",
            schema={
                "type": "array",
                "items": [{"type": "boolean"}, {"type": "integer"}, {"type": "string"}],
                "additionalItems": {"type": "boolean"},
            },
            data=[True, 1, "foo"],
            valid=True,
        ),
        DataTest(
            name="array, sequence with valid extra item",
            schema={
                "type": "array",
                "items": [{"type": "boolean"}, {"type": "integer"}, {"type": "string"}],
                "additionalItems": {"type": "boolean"},
            },
            data=[True, 1, "foo", True],
            valid=True,
        ),
        DataTest(
            name="array, sequence with invalid extra item",
            schema={
                "type": "array",
                "items": [{"type": "boolean"}, {"type": "integer"}, {"type": "string"}],
                "additionalItems": {"type": "boolean"},
            },
            data=[True, 1, "foo", 1],
            valid=False,
            expected_errors={"3": [messages.INVALID_BOOLEAN]},
        ),
    ]


class AdditionalProperties(SchemaTest):
    schema_false = {
        "properties": {"foo": {}, "bar": {}},
        "patternProperties": {"^v": {}},
        "additionalProperties": False,
    }
    schema_allows_schema = {
        "properties": {"foo": {}, "bar": {}},
        "additionalProperties": {"type": "boolean"},
    }
    schema_alone = {"additionalProperties": {"type": "boolean"}}
    schema_no = {"properties": {"foo": {}, "bar": {}}}
    data_tests = [
        DataTest(
            name="schema_false, no additional properties is valid",
            schema=schema_false,
            data={"foo": 1},
            valid=True,
        ),
        DataTest(
            name="schema_false, an additional property is invalid",
            schema=schema_false,
            data={"foo": 1, "bar": 2, "quux": "boom"},
            valid=False,
            expected_errors={
                # Used with jsonschema 2.4.0:
                # '': [messages.UNEXPECTED_FIELD],
                # This changed in jsonschema 2.6.0:
                # https://github.com/Julian/jsonschema/pull/317
                "": [messages.REGEX_NOT_MATCHED]
            },
        ),
        DataTest(
            name="schema_false, ignores non-objects",
            schema=schema_false,
            data=[1, 2, 3],
            valid=True,
        ),
        DataTest(
            name="schema_false, patternProperties are not additional properties",
            schema=schema_false,
            data={"foo": 1, "vroom": 2},
            valid=True,
        ),
        DataTest(
            name="schema_allows_schema, no additional properties",
            schema=schema_allows_schema,
            data={"foo": 1},
            valid=True,
        ),
        DataTest(
            name="schema_allows_schema, additional valid property",
            schema=schema_allows_schema,
            data={"foo": 1, "bar": 2, "quux": True},
            valid=True,
        ),
        DataTest(
            name="schema_allows_schema, additional invalid property",
            schema=schema_allows_schema,
            data={"foo": 1, "bar": 2, "quux": 12},
            valid=False,
            expected_errors={"quux": [messages.INVALID_BOOLEAN]},
        ),
        DataTest(
            name="schema_alone, additional valid property",
            schema=schema_alone,
            data={"foo": True},
            valid=True,
        ),
        DataTest(
            name="schema_alone, additional invalid property",
            schema=schema_alone,
            data={"foo": 1},
            valid=False,
            expected_errors={"foo": [messages.INVALID_BOOLEAN]},
        ),
        DataTest(
            name="schema_no, additional properties are allowed",
            schema=schema_no,
            data={"foo": 1, "bar": 2, "quux": True},
            valid=True,
        ),
    ]


class AllOf(SchemaTest):
    schema = {
        "type": "object",
        "properties": {"A": {"type": "boolean"}, "B": {"type": "boolean"}},
        "allOf": [{"required": ["A"]}, {"required": ["B"]}],
    }
    data_tests = [
        DataTest(
            name="Empty object",
            data={},
            valid=False,
            expected_errors={
                "A": [messages.MISSING_REQUIRED_FIELD],
                "B": [messages.MISSING_REQUIRED_FIELD],
            },
        ),
        DataTest(
            name="Mismatch B",
            data={"A": True},
            valid=False,
            expected_errors={"B": [messages.MISSING_REQUIRED_FIELD]},
        ),
        DataTest(
            name="Mismatch A",
            data={"B": True},
            valid=False,
            expected_errors={"A": [messages.MISSING_REQUIRED_FIELD]},
        ),
        DataTest(name="Both values", data={"A": False, "B": False}, valid=True),
        DataTest(
            name="Wrong type for A",
            data={"A": 1, "B": False},
            valid=False,
            expected_errors={"A": [messages.INVALID_BOOLEAN]},
        ),
        DataTest(
            name="Wrong type for both",
            data={"A": 1, "B": 1},
            valid=False,
            expected_errors={
                "A": [messages.INVALID_BOOLEAN],
                "B": [messages.INVALID_BOOLEAN],
            },
        ),
    ]


class AnyOf(SchemaTest):
    schema = {
        "type": "object",
        "properties": {"A": {"type": "boolean"}, "B": {"type": "boolean"}},
        "anyOf": [{"required": ["A"]}, {"required": ["B"]}],
    }
    data_tests = [
        DataTest(
            name="Empty object",
            data={},
            valid=False,
            expected_errors={"": [messages.NOT_VALID_UNDER_ANY_SCHEMA]},
        ),
        DataTest(name="Mismatch B", data={"A": True}, valid=True),
        DataTest(name="Mismatch A", data={"B": True}, valid=True),
        DataTest(name="Both values", data={"A": False, "B": False}, valid=True),
        DataTest(
            name="Wrong type for A",
            data={"A": 1, "B": False},
            valid=False,
            expected_errors={"A": [messages.INVALID_BOOLEAN]},
        ),
        DataTest(
            name="Wrong type for both",
            data={"A": 1, "B": 1},
            valid=False,
            expected_errors={
                "A": [messages.INVALID_BOOLEAN],
                "B": [messages.INVALID_BOOLEAN],
            },
        ),
    ]


class Dependencies(SchemaTest):
    schema_single = {"dependencies": {"bar": ["foo"]}}
    schema_multiple = {"dependencies": {"quux": ["foo", "bar"]}}
    schema_multiple_subschema = {
        "dependencies": {
            "bar": {
                "properties": {"foo": {"type": "integer"}, "bar": {"type": "integer"}}
            }
        }
    }
    data_tests = [
        DataTest(name="single, neither", schema=schema_single, data={}, valid=True),
        DataTest(
            name="single, nondependant",
            schema=schema_single,
            data={"foo": 1},
            valid=True,
        ),
        DataTest(
            name="single, with dependency",
            schema=schema_single,
            data={"foo": 1, "bar": 2},
            valid=True,
        ),
        DataTest(
            name="single, missing dependency",
            schema=schema_single,
            data={"bar": 2},
            valid=False,
            expected_errors={"": [messages.MISSING_DEPENDENT_FIELD]},
        ),
        DataTest(
            name="single, ignores non-objects",
            schema=schema_single,
            data="foo",
            valid=True,
        ),
        DataTest(name="multiple, neither", schema=schema_multiple, data={}, valid=True),
        DataTest(
            name="multiple, nondependants",
            schema=schema_multiple,
            data={"foo": 1, "bar": 2},
            valid=True,
        ),
        DataTest(
            name="multiple, with dependencies",
            schema=schema_multiple,
            data={"foo": 1, "bar": 2, "quux": 3},
            valid=True,
        ),
        DataTest(
            name="multiple, missing dependency",
            schema=schema_multiple,
            data={"foo": 1, "quux": 2},
            valid=False,
            expected_errors={"": [messages.MISSING_DEPENDENT_FIELD]},
        ),
        DataTest(
            name="multiple, missing other dependency",
            schema=schema_multiple,
            data={"bar": 1, "quux": 2},
            valid=False,
            expected_errors={"": [messages.MISSING_DEPENDENT_FIELD]},
        ),
        DataTest(
            name="multiple, missing both dependencies",
            schema=schema_multiple,
            data={"quux": 1},
            valid=False,
            expected_errors={
                "": [messages.MISSING_DEPENDENT_FIELD, messages.MISSING_DEPENDENT_FIELD]
            },
        ),
        DataTest(
            name="multiple subschema, valid",
            schema=schema_multiple_subschema,
            data={"foo": 1, "bar": 2},
            valid=True,
        ),
        DataTest(
            name="multiple subschema, no dependency",
            schema=schema_multiple_subschema,
            data={"foo": "quux"},
            valid=True,
        ),
        DataTest(
            name="multiple subschema, wrong type",
            schema=schema_multiple_subschema,
            data={"foo": "quux", "bar": 2},
            valid=False,
            expected_errors={"foo": [messages.INVALID_INT]},
        ),
        DataTest(
            name="multiple subschema, wrong type other",
            schema=schema_multiple_subschema,
            data={"foo": 2, "bar": "quux"},
            valid=False,
            expected_errors={"bar": [messages.INVALID_INT]},
        ),
        DataTest(
            name="multiple subschema, wrong type both",
            schema=schema_multiple_subschema,
            data={"foo": "quux", "bar": "quux"},
            valid=False,
            expected_errors={
                "foo": [messages.INVALID_INT],
                "bar": [messages.INVALID_INT],
            },
        ),
    ]


class Enum(SchemaTest):
    schema_simple = {"enum": [1, 2, 3]}
    schema_heterogeneous = {"enum": [6, "foo", [], True, {"foo": 12}]}
    schema_properties = {
        "type": "object",
        "properties": {"foo": {"enum": ["foo"]}, "bar": {"enum": ["bar"]}},
        "required": ["bar"],
    }
    data_tests = [
        DataTest(name="simple, valid", schema=schema_simple, data=1, valid=True),
        DataTest(
            name="simple, invalid",
            schema=schema_simple,
            data=4,
            valid=False,
            expected_errors={"": [messages.VALUE_NOT_IN_CHOICES]},
        ),
        DataTest(
            name="heterogeneous, valid",
            schema=schema_heterogeneous,
            data=[],
            valid=True,
        ),
        DataTest(
            name="heterogeneous, invalid",
            schema=schema_heterogeneous,
            data=None,
            valid=False,
            expected_errors={"": [messages.VALUE_NOT_IN_CHOICES]},
        ),
        DataTest(
            name="heterogeneous, deep valid",
            schema=schema_heterogeneous,
            data={"foo": False},
            valid=False,
            expected_errors={"": [messages.VALUE_NOT_IN_CHOICES]},
        ),
        DataTest(
            name="properties, both valid",
            schema=schema_properties,
            data={"foo": "foo", "bar": "bar"},
            valid=True,
        ),
        DataTest(
            name="properties, missing optional valid",
            schema=schema_properties,
            data={"bar": "bar"},
            valid=True,
        ),
        DataTest(
            name="properties, missing required",
            schema=schema_properties,
            data={"foo": "foo"},
            valid=False,
            expected_errors={"bar": [messages.MISSING_REQUIRED_FIELD]},
        ),
        DataTest(
            name="properties, missing all",
            schema=schema_properties,
            data={},
            valid=False,
            expected_errors={"bar": [messages.MISSING_REQUIRED_FIELD]},
        ),
    ]


class Format(SchemaTest):
    schema = {
        "type": "object",
        "properties": {
            "datetimes": {
                "type": "array",
                "items": {"type": "string", "format": "date-time"},
            },
            "emails": {"type": "array", "items": {"type": "string", "format": "email"}},
            "ipv4s": {"type": "array", "items": {"type": "string", "format": "ipv4"}},
            "ipv6s": {"type": "array", "items": {"type": "string", "format": "ipv6"}},
            "hostnames": {
                "type": "array",
                "items": {"type": "string", "format": "hostname"},
            },
            "urls": {"type": "array", "items": {"type": "string", "format": "url"}},
            "regexes": {
                "type": "array",
                "items": {"type": "string", "format": "regex"},
            },
        },
    }
    data_tests = [
        DataTest(
            name="datetime. valid",
            data={
                "datetimes": ["2013-03-25T12:42:31+00:32", "2013-03-25T22:04:10.04399Z"]
            },
            valid=True,
        ),
        DataTest(
            name="datetime. invalid",
            data={
                "datetimes": [
                    "2015-05-13 13:35:15.718978",
                    "2015-05-13 13:35:15",
                    "13-05-2013",
                ]
            },
            valid=False,
            expected_errors={
                "datetimes.0": [messages.INVALID_DATETIME],
                "datetimes.1": [messages.INVALID_DATETIME],
                "datetimes.2": [messages.INVALID_DATETIME],
            },
        ),
        DataTest(
            name="email. valid",
            data={
                "emails": [
                    "johndoe@domain.com",
                    "john.doe@domain.com",
                    "john.doe@sub.domain.com",
                    "j@sub.domain.com",
                    "j@d.com",
                    "j@domain.co.uk",
                ]
            },
            valid=True,
        ),
        DataTest(
            name="email. invalid",
            data={
                "emails": [
                    "",
                    "johndoe",
                    "johndoe@domain",
                    "johndoe@domain.",
                    "@domain",
                    "@domain.com",
                    "domain.com",
                ]
            },
            valid=False,
            expected_errors={
                "emails.0": [messages.INVALID_EMAIL],
                "emails.1": [messages.INVALID_EMAIL],
                "emails.2": [messages.INVALID_EMAIL],
                "emails.3": [messages.INVALID_EMAIL],
                "emails.4": [messages.INVALID_EMAIL],
                "emails.5": [messages.INVALID_EMAIL],
                "emails.6": [messages.INVALID_EMAIL],
            },
        ),
        DataTest(
            name="ipv4. valid",
            data={
                "ipv4s": [
                    "98.139.180.149",
                    "69.89.31.226",
                    "192.168.1.1",
                    "127.0.0.0",
                    "0.0.0.0",
                    "255.255.255.255",
                ]
            },
            valid=True,
        ),
        DataTest(
            name="ipv4. invalid",
            data={
                "ipv4s": [
                    "",
                    "0",
                    "0.",
                    "0.0",
                    "0.0.",
                    "0.0.0",
                    "0.0.0.0.",
                    "0.0.0.0.0",
                    "256.256.256.256",
                    "2002:4559:1FE2::4559:1FE2",
                    "2002:4559:1FE2:0:0:0:4559:1FE2",
                    "2002:4559:1FE2:0000:0000:0000:4559:1FE2",
                ]
            },
            valid=False,
            expected_errors={
                "ipv4s.0": [messages.INVALID_IPV4],
                "ipv4s.1": [messages.INVALID_IPV4],
                "ipv4s.2": [messages.INVALID_IPV4],
                "ipv4s.3": [messages.INVALID_IPV4],
                "ipv4s.4": [messages.INVALID_IPV4],
                "ipv4s.5": [messages.INVALID_IPV4],
                "ipv4s.6": [messages.INVALID_IPV4],
                "ipv4s.7": [messages.INVALID_IPV4],
                "ipv4s.8": [messages.INVALID_IPV4],
                "ipv4s.9": [messages.INVALID_IPV4],
                "ipv4s.10": [messages.INVALID_IPV4],
                "ipv4s.11": [messages.INVALID_IPV4],
            },
        ),
        DataTest(
            name="ipv6. valid",
            data={
                "ipv6s": [
                    "2002:4559:1FE2::4559:1FE2",
                    "2002:4559:1FE2:0:0:0:4559:1FE2",
                    "2002:4559:1FE2:0000:0000:0000:4559:1FE2",
                ]
            },
            valid=True,
        ),
        DataTest(
            name="ipv6. invalid",
            data={
                "ipv6s": [
                    "",
                    "98.139.180.149",
                    "69.89.31.226",
                    "192.168.1.1",
                    "127.0.0.0",
                    "0.0.0.0",
                    "255.255.255.255",
                ]
            },
            valid=False,
            expected_errors={
                "ipv6s.0": [messages.INVALID_IPV6],
                "ipv6s.1": [messages.INVALID_IPV6],
                "ipv6s.2": [messages.INVALID_IPV6],
                "ipv6s.3": [messages.INVALID_IPV6],
                "ipv6s.4": [messages.INVALID_IPV6],
                "ipv6s.5": [messages.INVALID_IPV6],
                "ipv6s.6": [messages.INVALID_IPV6],
            },
        ),
        DataTest(
            name="hostname. valid",
            data={
                "hostnames": [
                    "localhost",
                    "google",
                    "google.com",
                    "xn--hxajbheg2az3al.xn--jxalpdlp",
                    "a" * 63,
                ]
            },
            valid=True,
        ),
        DataTest(
            name="hostname. invalid",
            data={"hostnames": ["...", "a" * 64, "-hi-", "_hi_", "*hi*"]},
            valid=False,
            expected_errors={
                "hostnames.0": [messages.INVALID_HOSTNAME],
                "hostnames.1": [messages.INVALID_HOSTNAME],
                "hostnames.2": [messages.INVALID_HOSTNAME],
                "hostnames.3": [messages.INVALID_HOSTNAME],
                "hostnames.4": [messages.INVALID_HOSTNAME],
            },
        ),
        DataTest(
            name="urls. valid",
            data={
                "urls": [
                    "http://www.domain",
                    "http://www.com",
                    "http://www.domain.com.",
                    "http://www.domain.com/.",
                    "http://www.domain.com/..",
                    "http://www.domain.com//cataglog//index.html",
                    "http://www.domain.net/",
                    "http://www.domain.com/level2/leafnode-L2.xhtml/",
                    "http://www.domain.com/level2/level3/leafnode-L3.xhtml/",
                    "http://www.domain.com?pageid=123&testid=1524",
                    "http://www.domain.com/do.html#A",
                ]
            },
            valid=True,
        ),
        DataTest(
            name="urls. invalid",
            data={
                "urls": [
                    "",
                    "http://",
                    "http://www.",
                    "www.",
                    "http://www. .com",
                    "domain.com",
                    "www.domain.com",
                    "http:/www.domain.com",
                    "http//www.domain.com",
                    "http:www.domain.com",
                    "htp://domain.com/",
                    "http://sub.domain.com\\en-us\\default.aspx\\",
                    "http:\\\\msdn.domain.com\\en-us\\library\\default.aspx\\",
                    "http:\\\\www.domain.com\\leafnode-L1.html",
                    "./",
                    "../",
                    "http:\\\\www.domain.com\\leafnode-L1.xhtml\\",
                ]
            },
            valid=False,
            expected_errors={
                "urls.0": [messages.INVALID_URL],
                "urls.1": [messages.INVALID_URL],
                "urls.2": [messages.INVALID_URL],
                "urls.3": [messages.INVALID_URL],
                "urls.4": [messages.INVALID_URL],
                "urls.5": [messages.INVALID_URL],
                "urls.6": [messages.INVALID_URL],
                "urls.7": [messages.INVALID_URL],
                "urls.8": [messages.INVALID_URL],
                "urls.9": [messages.INVALID_URL],
                "urls.10": [messages.INVALID_URL],
                "urls.11": [messages.INVALID_URL],
                "urls.12": [messages.INVALID_URL],
                "urls.13": [messages.INVALID_URL],
                "urls.14": [messages.INVALID_URL],
                "urls.15": [messages.INVALID_URL],
                "urls.16": [messages.INVALID_URL],
            },
        ),
        DataTest(name="regexes. valid", data={"regexes": ["[0-9]+"]}, valid=True),
        DataTest(
            name="regexes. invalid",
            data={"regexes": ["[0-9]++"]},
            valid=False,
            expected_errors={"regexes.0": [messages.INVALID_REGEX]},
        ),
    ]


class Items(SchemaTest):
    schema_items = {"items": {"type": "integer"}}
    schema_array = {"items": [{"type": "integer"}, {"type": "string"}]}
    data_tests = [
        DataTest(
            name="schema_items. valid items",
            schema=schema_items,
            data=[1, 2, 3],
            valid=True,
        ),
        DataTest(
            name="schema_items. wrong type of items",
            schema=schema_items,
            data=[1, "x"],
            valid=False,
            expected_errors={"1": [messages.INVALID_INT]},
        ),
        DataTest(
            name="schema_items. ignores non-arrays",
            schema=schema_items,
            data={"foo": "bar"},
            valid=True,
        ),
        DataTest(
            name="schema_array. correct types",
            schema=schema_array,
            data=[1, "foo"],
            valid=True,
        ),
        DataTest(
            name="schema_array. wrong types",
            schema=schema_array,
            data=["foo", 1],
            valid=False,
            expected_errors={
                "0": [messages.INVALID_INT],
                "1": [messages.INVALID_STRING],
            },
        ),
    ]


class MaxItems(SchemaTest):
    schema = {"maxItems": 2}
    data_tests = [
        DataTest(name="shorter is valid", data=[1], valid=True),
        DataTest(name="exact length is valid", data=[1, 2], valid=True),
        DataTest(
            name="too long is invalid",
            data=[1, 2, 3],
            valid=False,
            expected_errors={"": [messages.FIELD_TOO_LONG]},
        ),
        DataTest(name="ignores non-arrays", data="foobar", valid=True),
    ]


class MaxLength(SchemaTest):
    schema = {"maxLength": 2}
    data_tests = [
        DataTest(name="shorter is valid", data="f", valid=True),
        DataTest(name="exact length is valid", data="fo", valid=True),
        DataTest(
            name="too long is invalid",
            data="foo",
            valid=False,
            expected_errors={"": [messages.FIELD_TOO_LONG]},
        ),
        DataTest(name="ignores non-strings", data=100, valid=True),
    ]


class MaxProperties(SchemaTest):
    schema = {"maxProperties": 2}
    data_tests = [
        DataTest(name="shorter is valid", data={"foo": 1}, valid=True),
        DataTest(name="exact length is valid", data={"foo": 1, "bar": 2}, valid=True),
        DataTest(
            name="too long is invalid",
            data={"foo": 1, "bar": 2, "baz": 3},
            valid=False,
            expected_errors={"": [messages.TOO_MANY_PROPERTIES]},
        ),
        DataTest(name="ignores non-objects", data="foobar", valid=True),
    ]


class Maximum(SchemaTest):
    # exclusiveMaximum behaviour changed from draft-04 to draft-06
    # http://json-schema.org/draft-06/json-schema-release-notes.html#backwards-incompatible-changes
    schema = {"maximum": 3.0}
    draft4_schema_exclusive = {
        "$schema": "http://json-schema.org/draft-04/schema",
        "maximum": 3.0,
        "exclusiveMaximum": True,
    }
    draft6_schema_exclusive = {
        "$schema": "http://json-schema.org/draft-06/schema",
        "exclusiveMaximum": 3.0,
    }

    data_tests = [
        DataTest(name="below", data=2.6, valid=True),
        DataTest(
            name="above",
            data=3.5,
            valid=False,
            expected_errors={"": [messages.NUMBER_TOO_HIGH]},
        ),
        DataTest(
            name="ignores non-numbers", data={"foo": 1, "bar": 2, "baz": 3}, valid=True
        ),
        DataTest(
            name="draft4 exclusive. below",
            schema=draft4_schema_exclusive,
            data=2.2,
            valid=True,
        ),
        DataTest(
            name="draft4_exclusive. boundary point",
            schema=draft4_schema_exclusive,
            data=3.0,
            valid=False,
            expected_errors={"": [messages.NUMBER_TOO_HIGH]},
        ),
        DataTest(
            name="draft4_exclusive. above",
            schema=draft4_schema_exclusive,
            data=3.5,
            valid=False,
            expected_errors={"": [messages.NUMBER_TOO_HIGH]},
        ),
        DataTest(
            name="draft6 exclusive. below",
            schema=draft6_schema_exclusive,
            data=2.2,
            valid=True,
        ),
        DataTest(
            name="draft6_exclusive. boundary point",
            schema=draft6_schema_exclusive,
            data=3.0,
            valid=False,
            expected_errors={"": [messages.NUMBER_TOO_HIGH]},
        ),
        DataTest(
            name="draft6_exclusive. above",
            schema=draft6_schema_exclusive,
            data=3.5,
            valid=False,
            expected_errors={"": [messages.NUMBER_TOO_HIGH]},
        ),
    ]


class MinItems(SchemaTest):
    schema = {"minItems": 1}
    data_tests = [
        DataTest(name="longer is valid", data=[1, 2], valid=True),
        DataTest(name="exact length is valid", data=[1], valid=True),
        DataTest(
            name="too short is invalid",
            data=[],
            valid=False,
            expected_errors={"": [messages.FIELD_TOO_SHORT]},
        ),
        DataTest(name="ignores non-arrays", data="", valid=True),
    ]


class MinProperties(SchemaTest):
    schema = {"minProperties": 1}
    data_tests = [
        DataTest(name="longer is valid", data={"foo": 1, "bar": 2}, valid=True),
        DataTest(name="exact length is valid", data={"foo": 1}, valid=True),
        DataTest(
            name="too short is invalid",
            data={},
            valid=False,
            expected_errors={"": [messages.NOT_ENOUGH_PROPERTIES]},
        ),
        DataTest(name="ignores non-objects", data="", valid=True),
    ]


class Minimum(SchemaTest):
    # exclusiveMinimum behaviour changed from draft-04 to draft-06
    # http://json-schema.org/draft-06/json-schema-release-notes.html#backwards-incompatible-changes
    schema = {"minimum": 1.1}
    draft4_schema_exclusive = {
        "$schema": "http://json-schema.org/draft-04/schema",
        "minimum": 1.1,
        "exclusiveMinimum": True,
    }
    draft6_schema_exclusive = {
        "$schema": "http://json-schema.org/draft-06/schema",
        "exclusiveMinimum": 1.1,
    }

    data_tests = [
        DataTest(name="above", data=2.6, valid=True),
        DataTest(
            name="below",
            data=0.6,
            valid=False,
            expected_errors={"": [messages.NUMBER_TOO_LOW]},
        ),
        DataTest(name="ignores non-numbers", data="x", valid=True),
        DataTest(
            name="exclusive. above",
            schema=draft4_schema_exclusive,
            data=1.2,
            valid=True,
        ),
        DataTest(
            name="exclusive. above",
            schema=draft6_schema_exclusive,
            data=1.2,
            valid=True,
        ),
        DataTest(
            name="draft4 exclusive. boundary point",
            schema=draft4_schema_exclusive,
            data=1.1,
            valid=False,
            expected_errors={"": [messages.NUMBER_TOO_LOW]},
        ),
        DataTest(
            name="draft6 exclusive. boundary point",
            schema=draft6_schema_exclusive,
            data=1.1,
            valid=False,
            expected_errors={"": [messages.NUMBER_TOO_LOW]},
        ),
        DataTest(
            name="draft4 exclusive. below",
            schema=draft4_schema_exclusive,
            data=0.6,
            valid=False,
            expected_errors={"": [messages.NUMBER_TOO_LOW]},
        ),
        DataTest(
            name="draft6 exclusive. below",
            schema=draft6_schema_exclusive,
            data=0.6,
            valid=False,
            expected_errors={"": [messages.NUMBER_TOO_LOW]},
        ),
    ]


class MultipleOf(SchemaTest):
    schema_int = {"multipleOf": 2}
    schema_number = {"multipleOf": 1.5}
    schema_small_number = {"multipleOf": 0.0001}
    data_tests = [
        DataTest(name="int. valid", schema=schema_int, data=10, valid=True),
        DataTest(name="int. valid float", schema=schema_int, data=10.0, valid=True),
        DataTest(
            name="int. invalid",
            schema=schema_int,
            data=7,
            valid=False,
            expected_errors={"": [messages.NOT_MULTIPLE_OF]},
        ),
        DataTest(name="int. ignores non-numbers", data="x", valid=True),
        DataTest(
            name="number. zero is multiple of anything",
            schema=schema_number,
            data=0,
            valid=True,
        ),
        DataTest(name="number. valid", schema=schema_number, data=4.5, valid=True),
        DataTest(
            name="number. invalid",
            schema=schema_number,
            data=35,
            valid=False,
            expected_errors={"": [messages.NOT_MULTIPLE_OF]},
        ),
        DataTest(
            name="small number. valid",
            schema=schema_small_number,
            data=0.0075,
            valid=True,
        ),
        DataTest(
            name="small number. invalid",
            schema=schema_small_number,
            data=0.00751,
            valid=False,
            expected_errors={"": [messages.NOT_MULTIPLE_OF]},
        ),
    ]


class Not(SchemaTest):
    schema_not = {"not": {"type": "integer"}}
    schema_multiple = {"not": {"type": ["integer", "boolean"]}}
    schema_complex = {
        "not": {"type": "object", "properties": {"foo": {"type": "string"}}}
    }
    schema_forbidden = {"properties": {"foo": {"not": {}}}}
    data_tests = [
        DataTest(name="not. allowed", schema=schema_not, data="foo", valid=True),
        DataTest(
            name="not. disallowed",
            schema=schema_not,
            data=1,
            valid=False,
            expected_errors={"": [messages.NOT_ALLOWED_VALUE]},
        ),
        DataTest(
            name="multiple. allowed", schema=schema_multiple, data="foo", valid=True
        ),
        DataTest(
            name="multiple. mismatch",
            schema=schema_multiple,
            data=1,
            valid=False,
            expected_errors={"": [messages.NOT_ALLOWED_VALUE]},
        ),
        DataTest(
            name="multiple. other mismatch",
            schema=schema_multiple,
            data=True,
            valid=False,
            expected_errors={"": [messages.NOT_ALLOWED_VALUE]},
        ),
        DataTest(name="complex. match", schema=schema_complex, data=1, valid=True),
        DataTest(
            name="complex. other match",
            schema=schema_complex,
            data={"foo": 1},
            valid=True,
        ),
        DataTest(
            name="complex. mismatch",
            schema=schema_complex,
            data={"foo": "bar"},
            valid=False,
            expected_errors={"": [messages.NOT_ALLOWED_VALUE]},
        ),
        DataTest(
            name="forbidden. present",
            schema=schema_forbidden,
            data={"foo": 1, "bar": 2},
            valid=False,
            expected_errors={"foo": [messages.NOT_ALLOWED_VALUE]},
        ),
        DataTest(
            name="forbidden. absent",
            schema=schema_forbidden,
            data={"bar": 1, "baz": 2},
            valid=True,
        ),
    ]


class OneOf(SchemaTest):
    schema = {
        "type": "object",
        "properties": {"A": {"type": "boolean"}, "B": {"type": "boolean"}},
        "oneOf": [{"required": ["A"]}, {"required": ["B"]}],
    }
    data_tests = [
        DataTest(
            name="Empty object",
            data={},
            valid=False,
            expected_errors={"": [messages.NOT_VALID_UNDER_ANY_SCHEMA]},
        ),
        DataTest(name="Mismatch B", data={"A": True}, valid=True),
        DataTest(name="Mismatch A", data={"B": True}, valid=True),
        DataTest(
            name="Both values",
            data={"A": False, "B": False},
            valid=False,
            expected_errors={"": [messages.VALID_FOR_SEVERAL_EXCLUSIVE_SCHEMAS]},
        ),
        DataTest(
            name="Wrong type for A",
            data={"A": 1, "B": False},
            valid=False,
            expected_errors={
                "": [messages.VALID_FOR_SEVERAL_EXCLUSIVE_SCHEMAS],
                "A": [messages.INVALID_BOOLEAN],
            },
        ),
        DataTest(
            name="Wrong type for both",
            data={"A": 1, "B": 1},
            valid=False,
            expected_errors={
                "": [messages.VALID_FOR_SEVERAL_EXCLUSIVE_SCHEMAS],
                "A": [messages.INVALID_BOOLEAN],
                "B": [messages.INVALID_BOOLEAN],
            },
        ),
    ]


class Pattern(SchemaTest):
    schema = {"pattern": "^a*$"}
    data_tests = [
        DataTest(name="valid", data="aaa", valid=True),
        DataTest(
            name="invalid",
            data="abc",
            valid=False,
            expected_errors={"": [messages.REGEX_NOT_MATCHED]},
        ),
        DataTest(name="ignores non-strings", data=True, valid=True),
    ]


class PatternProperties(SchemaTest):
    schema_single = {"patternProperties": {"f.*o": {"type": "integer"}}}
    schema_multiple = {
        "patternProperties": {"a*": {"type": "integer"}, "aaa*": {"maximum": 20}}
    }
    schema_complex = {
        "patternProperties": {
            "[0-9]{2,}": {"type": "boolean"},
            "X_": {"type": "string"},
        }
    }
    data_tests = [
        DataTest(
            name="single. valid", schema=schema_single, data={"foo": 1}, valid=True
        ),
        DataTest(
            name="single. multiple valid",
            schema=schema_single,
            data={"foo": 1, "foooooo": 2},
            valid=True,
        ),
        DataTest(
            name="single. invalid",
            schema=schema_single,
            data={"foo": "bar", "fooooo": 2},
            valid=False,
            expected_errors={"foo": [messages.INVALID_INT]},
        ),
        DataTest(
            name="single. multiple invalid",
            schema=schema_single,
            data={"foo": "bar", "foooooo": "baz"},
            valid=False,
            expected_errors={
                "foo": [messages.INVALID_INT],
                "foooooo": [messages.INVALID_INT],
            },
        ),
        DataTest(
            name="single. ignores non-objects",
            schema=schema_single,
            data=12,
            valid=True,
        ),
        DataTest(
            name="multiple. valid", schema=schema_multiple, data={"a": 21}, valid=True
        ),
        DataTest(
            name="multiple. simultaneous valid",
            schema=schema_multiple,
            data={"aaaa": 18},
            valid=True,
        ),
        DataTest(
            name="multiple. multiple valid",
            schema=schema_multiple,
            data={"a": 21, "aaaa": 18},
            valid=True,
        ),
        DataTest(
            name="multiple. invalid one",
            schema=schema_multiple,
            data={"a": "bar"},
            valid=False,
            expected_errors={"a": [messages.INVALID_INT]},
        ),
        DataTest(
            name="multiple. invalid other",
            schema=schema_multiple,
            data={"aaaa": 31},
            valid=False,
            expected_errors={"aaaa": [messages.NUMBER_TOO_HIGH]},
        ),
        DataTest(
            name="multiple. invalid both",
            schema=schema_multiple,
            data={"aaa": "foo", "aaaa": 31},
            valid=False,
            expected_errors={
                "aaa": [messages.INVALID_INT],
                "aaaa": [messages.NUMBER_TOO_HIGH],
            },
        ),
        DataTest(
            name="complex. non recognized members are ignored",
            schema=schema_complex,
            data={"answer 1": "42"},
            valid=True,
        ),
        DataTest(
            name="complex. recognized members are accounted for",
            schema=schema_complex,
            data={"a31b": None},
            valid=False,
            expected_errors={"a31b": [messages.INVALID_BOOLEAN]},
        ),
        DataTest(
            name="complex. regexes are case sensitive",
            schema=schema_complex,
            data={"a_x_3": 3},
            valid=True,
        ),
        DataTest(
            name="complex. regexes are case sensitive 2",
            schema=schema_complex,
            data={"a_X_3": 3},
            valid=False,
            expected_errors={"a_X_3": [messages.INVALID_STRING]},
        ),
    ]


class Properties(SchemaTest):
    schema = {"properties": {"foo": {"type": "integer"}, "bar": {"type": "string"}}}
    schema_mixed = {
        "properties": {
            "foo": {"type": "array", "maxItems": 3},
            "bar": {"type": "array"},
        },
        "patternProperties": {"f.o": {"minItems": 2}},
        "additionalProperties": {"type": "integer"},
    }
    data_tests = [
        DataTest(name="both present", data={"foo": 1, "bar": "baz"}, valid=True),
        DataTest(
            name="one invalid",
            data={"foo": 1, "bar": {}},
            valid=False,
            expected_errors={"bar": [messages.INVALID_STRING]},
        ),
        DataTest(
            name="both invalid",
            data={"foo": [], "bar": {}},
            valid=False,
            expected_errors={
                "foo": [messages.INVALID_INT],
                "bar": [messages.INVALID_STRING],
            },
        ),
        DataTest(name="other properties", data={"quux": []}, valid=True),
        DataTest(name="ignores non-objects", data=[], valid=True),
        DataTest(
            name="mixed. property validates property",
            schema=schema_mixed,
            data={"foo": [1, 2]},
            valid=True,
        ),
        DataTest(
            name="mixed. property invalidates property",
            schema=schema_mixed,
            data={"foo": [1, 2, 3, 4]},
            valid=False,
            expected_errors={"foo": [messages.FIELD_TOO_LONG]},
        ),
        DataTest(
            name="mixed. patternProperty invalidates property",
            schema=schema_mixed,
            data={"foo": []},
            valid=False,
            expected_errors={"foo": [messages.FIELD_TOO_SHORT]},
        ),
        DataTest(
            name="mixed. patternProperty validates nonproperty",
            schema=schema_mixed,
            data={"fxo": [1, 2]},
            valid=True,
        ),
        DataTest(
            name="mixed. patternProperty invalidates nonproperty",
            schema=schema_mixed,
            data={"fxo": []},
            valid=False,
            expected_errors={"fxo": [messages.FIELD_TOO_SHORT]},
        ),
        DataTest(
            name="mixed. additionalProperty ignores property",
            schema=schema_mixed,
            data={"bar": []},
            valid=True,
        ),
        DataTest(
            name="mixed. additionalProperty validates others",
            schema=schema_mixed,
            data={"quux": 3},
            valid=True,
        ),
        DataTest(
            name="mixed. additionalProperty invalidates others",
            schema=schema_mixed,
            data={"quux": "foo"},
            valid=False,
            expected_errors={"quux": [messages.INVALID_INT]},
        ),
    ]


class Ref(SchemaTest):
    schema_root = {"properties": {"foo": {"$ref": "#"}}, "additionalProperties": False}
    schema_relative = {
        "properties": {"foo": {"type": "integer"}, "bar": {"$ref": "#/properties/foo"}}
    }
    schema_relative_array = {"items": [{"type": "integer"}, {"$ref": "#/items/0"}]}
    schema_escaped = {
        "tilda~field": {"type": "integer"},
        "slash/field": {"type": "integer"},
        "percent%field": {"type": "integer"},
        "properties": {
            "tilda": {"$ref": "#/tilda~0field"},
            "slash": {"$ref": "#/slash~1field"},
            "percent": {"$ref": "#/percent%25field"},
        },
    }
    schema_nested = {
        "definitions": {
            "a": {"type": "integer"},
            "b": {"$ref": "#/definitions/a"},
            "c": {"$ref": "#/definitions/b"},
        },
        "$ref": "#/definitions/c",
    }
    data_tests = [
        DataTest(
            name="root. match", schema=schema_root, data={"foo": False}, valid=True
        ),
        DataTest(
            name="root. recursive match",
            schema=schema_root,
            data={"foo": {"foo": False}},
            valid=True,
        ),
        DataTest(
            name="root. mismatch",
            schema=schema_root,
            data={"bar": False},
            valid=False,
            expected_errors={
                "": [
                    messages.UNEXPECTED_FIELDS.format(
                        unexpected_fields="('bar' was unexpected)"
                    )
                ]
            },
        ),
        DataTest(
            name="root. recursive mismatch",
            schema=schema_root,
            data={"foo": {"bar": False}},
            valid=False,
            expected_errors={
                "foo": [
                    messages.UNEXPECTED_FIELDS.format(
                        unexpected_fields="('bar' was unexpected)"
                    )
                ]
            },
        ),
        DataTest(
            name="relative. match", schema=schema_relative, data={"bar": 3}, valid=True
        ),
        DataTest(
            name="relative. mismatch",
            schema=schema_relative,
            data={"bar": True},
            valid=False,
            expected_errors={"bar": [messages.INVALID_INT]},
        ),
        DataTest(
            name="relative array. match",
            schema=schema_relative_array,
            data=[1, 2],
            valid=True,
        ),
        DataTest(
            name="relative array. mismatch",
            schema=schema_relative_array,
            data=[1, "foo"],
            valid=False,
            expected_errors={"1": [messages.INVALID_INT]},
        ),
        DataTest(
            name="escaped. slash",
            schema=schema_escaped,
            data={"slash": "aoeu"},
            valid=False,
            expected_errors={"slash": [messages.INVALID_INT]},
        ),
        DataTest(
            name="escaped. tilda",
            schema=schema_escaped,
            data={"tilda": "aoeu"},
            valid=False,
            expected_errors={"tilda": [messages.INVALID_INT]},
        ),
        DataTest(
            name="escaped. percent",
            schema=schema_escaped,
            data={"percent": "aoeu"},
            valid=False,
            expected_errors={"percent": [messages.INVALID_INT]},
        ),
        DataTest(
            name="escaped. all",
            schema=schema_escaped,
            data={"slash": "aoeu", "tilda": "aoeu", "percent": "aoeu"},
            valid=False,
            expected_errors={
                "slash": [messages.INVALID_INT],
                "tilda": [messages.INVALID_INT],
                "percent": [messages.INVALID_INT],
            },
        ),
        DataTest(name="nested. valid", schema=schema_nested, data=5, valid=True),
        DataTest(
            name="nested. invalid",
            schema=schema_nested,
            data="a",
            valid=False,
            expected_errors={"": [messages.INVALID_INT]},
        ),
    ]


class Required(SchemaTest):
    schema = {"properties": {"foo": {}, "bar": {}}, "required": ["foo"]}
    schema_default = {"properties": {"foo": {}}}
    data_tests = [
        DataTest(name="valid", data={"foo": 1}, valid=True),
        DataTest(
            name="invalid",
            data={"bar": 1},
            valid=False,
            expected_errors={"foo": [messages.MISSING_REQUIRED_FIELD]},
        ),
        DataTest(
            name="not required by default", schema=schema_default, data={}, valid=True
        ),
    ]


class Type(SchemaTest):
    type_tests = [
        # -------------------------------------------------------
        # type          data        expected error
        # -------------------------------------------------------
        # array
        ("array", [], None),
        ("array", [1, 2], None),
        ("array", None, messages.INVALID_ARRAY),
        ("array", {}, messages.INVALID_ARRAY),
        ("array", "[]", messages.INVALID_ARRAY),
        ("array", "abc", messages.INVALID_ARRAY),
        ("array", 1, messages.INVALID_ARRAY),
        # boolean
        ("boolean", True, None),
        ("boolean", False, None),
        ("boolean", None, messages.INVALID_BOOLEAN),
        ("boolean", "True", messages.INVALID_BOOLEAN),
        ("boolean", "False", messages.INVALID_BOOLEAN),
        ("boolean", 0, messages.INVALID_BOOLEAN),
        ("boolean", 1, messages.INVALID_BOOLEAN),
        # integer
        ("integer", 0, None),
        ("integer", 1, None),
        ("integer", -500, None),
        ("integer", 1000000, None),
        ("integer", None, messages.INVALID_INT),
        ("integer", 1.2, messages.INVALID_INT),
        ("integer", "1", messages.INVALID_INT),
        # number
        ("number", 0, None),
        ("number", 1, None),
        ("number", -500, None),
        ("number", 1000000, None),
        ("number", 1.2, None),
        ("number", -34.7, None),
        ("number", None, messages.INVALID_NUMBER),
        ("number", "1", messages.INVALID_NUMBER),
        ("number", True, messages.INVALID_NUMBER),
        # null
        ("null", None, None),
        ("null", [], messages.NOT_NULL),
        ("null", {}, messages.NOT_NULL),
        ("null", 0, messages.NOT_NULL),
        ("null", "", messages.NOT_NULL),
        # object
        ("object", {}, None),
        ("object", {"a": 1}, None),
        ("object", None, messages.INVALID_OBJECT),
        ("object", [], messages.INVALID_OBJECT),
        ("object", 0, messages.INVALID_OBJECT),
        ("object", "", messages.INVALID_OBJECT),
        ("object", "abc", messages.INVALID_OBJECT),
        # string
        ("string", "", None),
        ("string", "abc", None),
        ("string", "-", None),
        ("string", "...", None),
        ("string", "", None),
        ("string", "abc", None),
        ("string", "España", None),
        ("string", "1", None),
        ("string", None, messages.INVALID_STRING),
        ("string", 1, messages.INVALID_STRING),
        ("string", [], messages.INVALID_STRING),
        ("string", {}, messages.INVALID_STRING),
    ]
    data_tests = [
        DataTest(
            name="%02d_%s" % (i + 1, data_type),
            data=data,
            valid=expected_error is None,
            expected_errors={"": [expected_error]} if expected_error else None,
            schema={"type": data_type},
        )
        for i, (data_type, data, expected_error) in enumerate(type_tests)
    ]


class Unique(SchemaTest):
    schema = {"uniqueItems": True}
    data_tests = [
        DataTest(name="unique array", data=[1, 2], valid=True),
        DataTest(
            name="non-unique array",
            data=[1, 1],
            valid=False,
            expected_errors={"": [messages.NOT_UNIQUE]},
        ),
        DataTest(
            name="non-unique numbers",
            data=[1.0, 1.00, 1],
            valid=False,
            expected_errors={"": [messages.NOT_UNIQUE]},
        ),
        DataTest(
            name="unique objects array",
            data=[{"foo": "bar"}, {"foo": "baz"}],
            valid=True,
        ),
        DataTest(
            name="non-unique objects array",
            data=[{"foo": "bar"}, {"foo": "bar"}],
            valid=False,
            expected_errors={"": [messages.NOT_UNIQUE]},
        ),
        DataTest(
            name="unique nested objects array",
            data=[{"foo": {"bar": {"baz": True}}}, {"foo": {"bar": {"baz": False}}}],
            valid=True,
        ),
        DataTest(
            name="non-unique nested objects array",
            data=[{"foo": {"bar": {"baz": True}}}, {"foo": {"bar": {"baz": True}}}],
            valid=False,
            expected_errors={"": [messages.NOT_UNIQUE]},
        ),
        DataTest(name="unique array of arrays", data=[["foo"], ["bar"]], valid=True),
        DataTest(
            name="non-unique array of arrays",
            data=[["foo"], ["foo"]],
            valid=False,
            expected_errors={"": [messages.NOT_UNIQUE]},
        ),
        DataTest(name="1 and True", data=[1, True], valid=True),
        DataTest(name="0 and False", data=[0, False], valid=True),
        DataTest(
            name="unique heterogeneous types", data=[{}, [1], True, None, 1], valid=True
        ),
        DataTest(
            name="non-unique heterogeneous types",
            data=[{}, [1], True, None, {}, 1],
            valid=False,
            expected_errors={"": [messages.NOT_UNIQUE]},
        ),
    ]



from spidermon.exceptions import NotConfigured


class JobMonitorMixin:
    @property
    def job(self):
        if not self.data.job:
            raise NotConfigured("Job not available!")
        return self.data.job



import pytest

from spidermon.contrib.validation.translator import MessageTranslator


@pytest.fixture
def message_translator():
    class FixtureMessageTranslator(MessageTranslator):
        messages = {
            r"Simple Message": "Translated Simple Message",
            r"^.+ is a required property$": "Missing Required Property",
            r"Options: (?P<options>.*)": "Translated With Options: {options}",
        }

    return FixtureMessageTranslator()


@pytest.mark.parametrize(
    "original_message,translated_message",
    [
        ("Simple Message", "Translated Simple Message"),
        ("email is a required property", "Missing Required Property"),
        ("Options: a, b, c", "Translated With Options: a, b, c"),
    ],
)
def test_message_translator(message_translator, original_message, translated_message):
    assert message_translator.translate_message(original_message) == translated_message
